#
# IS-IS transformation module
#
from box import Box

from . import _Module,_routing
from . import bfd
from .. import common
from ..augment import devices

class ISIS(_Module):

  def node_post_transform(self, node: Box, topology: Box) -> None:
    features = devices.get_device_features(node,topology.defaults)

    for af in ('ipv4','ipv6'):
      is_unnumbered = False
      for l in node.get('interfaces',[]):
        is_unnumbered = is_unnumbered or \
          'unnumbered' in l or \
          (af in l and isinstance(l[af],bool) and l[af])

      if is_unnumbered and not features.isis.unnumbered[af]:
        common.error(
          f'Device {node.device} used on node {node.name} cannot run IS-IS over {"unnumbered" if af == "ipv4" else "LLA"} {af} interfaces',
          common.IncorrectValue,
          'interfaces')

    for l in node.get('interfaces',[]):
      unnum_v4 = 'unnumbered' in l or ('ipv4' in l and isinstance(l.ipv4,bool) and l.ipv4)
      if unnum_v4 and \
          len(l.neighbors) > 1 and \
          features.isis.unnumbered.ipv4 and \
          not features.isis.unnumbered.network:
        common.error(
          f'Device {node.device} used on node {node.name} cannot run IS-IS over\n'+
          f'.. unnumbered multi-access interfaces (link {l.name})',
          common.IncorrectValue,
          'interfaces')
      elif _routing.external(l,'isis') or not (l.get('ipv4',False) or l.get('ipv6',False)):
        l.pop('isis',None) # Don't run IS-IS on external interfaces, or l2-only
      else:
        _routing.passive(l,'isis')
        err = _routing.network_type(l,'isis',['point-to-point'])
        if err:
          common.error(f'{err}\n... node {node.name} link {l}')

    #
    # Final steps:
    #
    # * remove IS-IS from VRF interfaces
    # * Calculate address families
    # * Enable BFD
    # * Remove ISIS module if there are no IS-IS enabled global interfaces
    #
    _routing.remove_vrf_interfaces(node,'isis')
    _routing.routing_af(node,'isis')
    bfd.multiprotocol_bfd_link_state(node,'isis')
    _routing.remove_unused_igp(node,'isis')


#
# netlab restart command
#
# Perform 'netlab down' followed by 'netlab up'
#
import typing
import argparse
import os
import glob
import subprocess
import shutil

from box import Box

from .. import common
from . import down, up
from . import common_parse_args

#
# Extra arguments for 'netlab up' command
#
def restart_parse_args() -> argparse.ArgumentParser:
  parser = argparse.ArgumentParser(
    description='Reconfigure and restart the virtual lab',
    parents = [ common_parse_args() ],
    add_help=True)
  parser.add_argument(
    '--no-config',
    dest='no_config',
    action='store_true',
    help='Do not configure lab devices')
  parser.add_argument(
    '--fast-config',
    dest='fast_config',
    action='store_true',
    help='Use fast device configuration (Ansible strategy = free)')
  parser.add_argument(
    '--snapshot',
    dest='snapshot',
    action='store',
    nargs='?',
    const='netlab.snapshot.yml',
    help='Use netlab snapshot file created by a previous lab run to start the lab')
  parser.add_argument(
    dest='topology', action='store', nargs='?',
    help='Topology file (default: topology.yml)')
  return parser

def run(cli_args: typing.List[str]) -> None:
  parser = restart_parse_args()
  args = parser.parse_args(cli_args)
  up_only_args = ['--fast-config','--no-config']

  down.run([ arg for arg in cli_args if arg not in up_only_args ])
  up.run(cli_args)



#
# Vagrant/libvirt provider module
#

from . import _Provider
from box import Box

class Virtualbox(_Provider):

  def transform_node_images(self, topology: Box) -> None:
    self.node_image_version(topology)



#
# Common routines for create-topology script
#
import sys
import typing
import warnings
import argparse
import os
import textwrap
import pathlib

from jinja2 import Environment, PackageLoader, StrictUndefined, make_logging_undefined
from box import Box,BoxList

LOGGING : bool = False
VERBOSE : int = 0
DEBUG : bool = False
QUIET : bool = False

RAISE_ON_ERROR : bool = False
WARNING : bool = False

AF_LIST = ['ipv4','ipv6']
BGP_SESSIONS = ['ibgp','ebgp']

err_count : int = 0
netsim_package_path = os.path.abspath(os.path.dirname(__file__))

class MissingValue(Warning):
  pass

class IncorrectValue(Warning):
  pass

class IncorrectType(Warning):
  pass

class FatalError(Warning):
  pass

class ErrorAbort(Exception):
  pass

def fatal(text: str, module: str = 'netsim-tools') -> None:
  global err_count
  err_count = err_count + 1
  if RAISE_ON_ERROR:
    raise ErrorAbort(text)
  else:
    if WARNING:
      warnings.warn_explicit(text,FatalError,filename=module,lineno=err_count)
    else:
      print(f'Fatal error in {module}: {text}',file=sys.stderr)
    sys.exit(1)

def error(text: str, category: typing.Type[Warning] = UserWarning, module: str = 'topology') -> None:
  global err_count
  err_count = err_count + 1
  if WARNING:
    warnings.warn_explicit(text,category,filename=module,lineno=err_count)
  else:
    print(f'{category.__name__} in {module}: {text}',file=sys.stderr)

def exit_on_error() -> None:
  global err_count
  if err_count > 0:
    fatal('Cannot proceed beyond this point due to errors, exiting')

def extra_data_printout(s : str) -> str:
  return textwrap.TextWrapper(
    initial_indent="... ",
    subsequent_indent="      ").fill(s)

#
# File functions
#

def open_output_file(fname: str) -> typing.TextIO:
  if fname == '-':
    return sys.stdout

  return open(fname,mode='w')

def close_output_file(f: typing.TextIO) -> None:
  if f.name != '<stdout>':
    f.close()

def print_yaml(x : typing.Any) -> str:
  if isinstance(x,dict):
    return Box(x).to_yaml()
  elif isinstance(x,list):
    return BoxList(x).to_yaml()
  else:
    return str(x)

def get_moddir() -> pathlib.Path:
  return pathlib.Path(__file__).resolve().parent

def template(j2: str , data: typing.Dict, path: str) -> str:
  ENV = Environment(loader=PackageLoader('netsim',path), \
          trim_blocks=True,lstrip_blocks=True, \
          undefined=make_logging_undefined(base=StrictUndefined))
  template = ENV.get_template(j2)
  return template.render(**data)

#
# Logging and debugging functions
#

def set_verbose(value: typing.Optional[int] = 1) -> None:
  global VERBOSE
  VERBOSE = 0 if value is None else value

def print_verbose(t: typing.Any) -> None:
  if VERBOSE:
    print(t)

#
# Sets common flags based on parsed arguments.
#
# Internal debugging flags (RAISE_ON_ERROR, WARNING) cannot be set with this function
#
def set_logging_flags(args: argparse.Namespace) -> None:
  global VERBOSE, LOGGING, DEBUG, QUIET, WARNING, RAISE_ON_ERROR
  
  if 'verbose' in args and args.verbose:
    VERBOSE = args.verbose

  if 'logging' in args and args.logging:
    LOGGING = True

  if 'debug' in args and args.debug:
    DEBUG = True

  if 'quiet' in args and args.quiet:
    QUIET = True

  if 'warning' in args and args.warning:
    WARNING = True

  if 'raise_on_error' in args and args.raise_on_error:
    RAISE_ON_ERROR = True

#
# Sets zero or more common flags and returns current settings.
#
# There must be a better way of doing it, but this is simple and it works
#
def set_flag(
      debug: typing.Optional[bool] = None,
      quiet: typing.Optional[bool] = None,
      verbose: typing.Optional[int] = None,
      logging: typing.Optional[bool] = None,
      warning: typing.Optional[bool] = None,
      raise_error: typing.Optional[bool] = None) -> dict:
  global DEBUG, VERBOSE, LOGGING, QUIET, RAISE_ON_ERROR, WARNING

  DEBUG = debug if debug is not None else DEBUG
  QUIET = quiet if quiet is not None else QUIET
  VERBOSE = verbose if verbose is not None else VERBOSE
  LOGGING = logging if logging is not None else LOGGING
  WARNING = warning if warning is not None else WARNING
  RAISE_ON_ERROR = raise_error  if raise_error is not None else RAISE_ON_ERROR

  return {
    'debug': DEBUG,
    'quiet': QUIET,
    'verbose': VERBOSE,
    'logging': LOGGING,
    'raise_on_error': RAISE_ON_ERROR,
    'warning': WARNING
  }



# Why use Graphite (https://github.com/netreplica/graphite) only with ContainerLab? :-)
#
# With this output module it would be possible to export a topology file to be used within Graphite.
# 
# Graphite container can be launched with:
#
# docker run -d \
#  -v "$(pwd)/graphite-default.json":/var/www/localhost/htdocs/default/default.json \
#  -p 8080:80 \
#  --name graphite \
#  netreplica/graphite:webssh2
#

from array import array
import typing

import json
import os
from box import Box

from .. import common
from .. import data
from . import _TopologyOutput

DEFAULT_NODE_ICON = "router"
HOST_BRIDGE_NODE_NAME = "({type}:{index})"

short_ifname_lookup = {
    'GigabitEthernet': 'ge',
    'Ethernet': 'eth',
    'ethernet': 'eth',
    'ether': 'eth'
}

def short_ifname(n : str) -> str:
    global short_ifname_lookup

    for long_name in short_ifname_lookup.keys():
        if long_name in n:
            return n.replace(long_name,short_ifname_lookup[long_name],1)

    return n

def nodes_items(topology: Box) -> list:
    r = []
    for name,n in topology.nodes.items():
        node_icon = ( data.get_from_box(n,'graphite.icon') or 
            data.get_from_box(topology,f'defaults.devices.{n.device}.graphite.icon') or 
            DEFAULT_NODE_ICON )
        graph_level = data.get_from_box(n,'graphite.level') or 1
        node_group = "tier-1"
        node_as = n.get('bgp', {}).get('as')
        if node_as:
            node_group = "as{}".format(node_as)
        r.append(
            {
                'name': name,
                'kind': n.device,
                "ipv4_address": n.mgmt.ipv4,
                "group": node_group,
                "labels": {
                    "graph-level": graph_level,
                    "graph-icon": node_icon,
                },
            }
        )
    # Special Case:
    # Create a fake node to identify a host bridge, in case there are 1 or more than 2 nodes attached to it.
    for l in topology.links:
        if (l.type == "lan" and l.node_count != 2) or l.type == "stub":
            # Create fake node
            # Inherit graph level and group from first node (l.interfaces[0].node)
            node_item = topology.nodes[l.interfaces[0].node]
            graph_level = data.get_from_box(node_item,'graphite.level') or 1
            node_group = "tier-1"
            node_as = node_item.get('bgp', {}).get('as')
            if node_as:
                node_group = "as{}".format(node_as)
            r.append(
                {
                    'name': HOST_BRIDGE_NODE_NAME.format(type=l.type, index=l.linkindex),
                    'kind': "(host bridge: {})".format(l.bridge),
                    "ipv4_address": '',
                    "group": node_group,
                    "labels": {
                        "graph-level": graph_level,
                        "graph-icon": 'expand',
                    },
                }
            )
    return r

def get_lan_intf_name(topology: Box, node_name: str, bridge_name: str) -> str:
    for intf in topology.nodes[node_name].interfaces:
        if intf.get('bridge','') == bridge_name:
            return short_ifname(intf.ifname)
    return "<unknown>"

def links_items(topology: Box) -> list:
    r = []
    for l in topology.links:
        # P2P Links
        if l.type == "p2p":
            r.append(
                {
                    "source": l.left.node,
                    "source_endpoint": short_ifname(l.left.ifname),
                    "target": l.right.node,
                    "target_endpoint": short_ifname(l.right.ifname)
                }
            )
        elif l.type == "lan" or l.type == "stub":
            # If LAN link have only two interfaces in it, treat it like a P2P
            # - Else, in nodes_items create a fake item that represents the host bridge, and attach the devices to it.
            if l.type == "lan" and l.node_count == 2:
                r.append(
                    {
                        "source": l.interfaces[0].node,
                        "source_endpoint": get_lan_intf_name(topology, l.interfaces[0].node, l.bridge),
                        "target": l.interfaces[1].node,
                        "target_endpoint": get_lan_intf_name(topology, l.interfaces[1].node, l.bridge),
                    }
                )
            else:
                bridge_node_name = HOST_BRIDGE_NODE_NAME.format(type=l.type, index=l.linkindex)
                for bridge_intf in l.interfaces:
                    # Create an attachment
                    r.append(
                        {
                            "source": bridge_node_name,
                            "source_endpoint": "",
                            "target": bridge_intf.node,
                            "target_endpoint": get_lan_intf_name(topology, bridge_intf.node, l.bridge),
                        }
                    )
    return r

class Graphite(_TopologyOutput):

  def write(self, topology: Box) -> None:
    outfile = self.settings.filename or 'graphite-default.json'

    graphite_json = {
        'nodes': nodes_items(topology),
        'links': links_items(topology),
    }

    output = common.open_output_file(outfile)

    output.write(json.dumps(graphite_json, indent=2,sort_keys=True))
    output.write("\n")

    if outfile != '-':
      common.close_output_file(output)
      print("Created Graphite topology file in %s" % outfile)
    else:
      output.write("\n")



#
# Create Ansible inventory
#
import typing

import yaml
import os
from box import Box

from .. import common
from . import _TopologyOutput

def node_with_label(f : typing.TextIO, n: Box, settings: Box, indent: typing.Optional[str] = '') -> None:
  f.write('%s  "%s" [\n' % (indent,n.name))
  f.write('%s    label=<%s [%s]<br /><sub>%s</sub>>\n' % (indent,n.name,n.device,n.loopback.ipv4 or n.loopback.ipv6 or ""))
  f.write('%s    fillcolor="%s"\n' % (indent,settings.colors.get('node','#ff9f01')))
  f.write('%s  ]\n' % indent)

def network_with_label(f : typing.TextIO, n: Box, settings: Box, indent: typing.Optional[str] = '') -> None:
  f.write('%s  "%s" [' % (indent,n.bridge))
  f.write('style=filled fillcolor="%s" fontsize=11' % (settings.colors.get("stub","#d1bfab")))
  f.write(' label="%s"' % (n.prefix.ipv4 or n.prefix.ipv6 or n.bridge))
  f.write("]\n")

def edge_label(f : typing.TextIO, direction: str, data: Box) -> None:
  addr = data.ipv4 or data.ipv6
  if addr:
    f.write(' %slabel="%s"' % (direction,addr))

def edge_p2p(f : typing.TextIO, l: Box, labels: typing.Optional[bool] = False) -> None:
  f.write(' "%s" -- "%s"' % (l.left.node, l.right.node))
  f.write(' [')
  if labels:
    edge_label(f,'tail',l[l.left.node])
    edge_label(f,'head',l[l.right.node])
  f.write(' ]\n')

def edge_node_net(f : typing.TextIO, l: Box, k: str, labels: typing.Optional[bool] = False) -> None:
  f.write(' "%s" -- "%s"' % (k,l.bridge))
  f.write(' [')
  if labels:
    edge_label(f,'tail',l[k])
  f.write(' ]\n')

def graph_start(f : typing.TextIO) -> None:
  f.write('graph {\n')
  f.write('  bgcolor="transparent"\n')
  f.write('  node [shape=box, style="rounded,filled" fontname=Verdana]\n')
  f.write('  edge [fontname=Verdana labelfontsize=10 labeldistance=1.5]\n')

def as_start(f : typing.TextIO, asn: str, label: typing.Optional[str], settings: Box) -> None:
  f.write('  subgraph cluster_%s {\n' % asn)
  f.write('    bgcolor="%s"\n' % settings.colors.get('as','#e8e8e8'))
  f.write('    fontname=Verdana\n')
  f.write('    margin=%s\n' % settings.margins.get('as',16))
  if label:
    f.write('    label="%s (AS %s)"\n' % (label,asn))
  else:
    f.write('    label="AS %s"\n' % asn)

def graph_bgp_clusters(f : typing.TextIO, bgp: Box, settings: Box) -> None:
  for asn in bgp.keys():
    as_start(f,asn,bgp[asn].get('name',None),settings)
    for n in bgp[asn].nodes.values():
      node_with_label(f,n,settings,'  ')
    f.write('  }\n')

def build_maps(topology: Box) -> Box:
  maps = Box({},default_box=True,box_dots=True)
  for name,n in topology.nodes.items():
    maps.nodes[name] = n

  if 'module' in topology and 'bgp' in topology.module:
    for name,n in topology.nodes.items():
      if 'bgp' in n and 'as' in n.bgp:
        maps.bgp[n.bgp['as']].nodes[n.name] = n

  if 'bgp' in topology and 'as_list' in topology.bgp:
    for (asn,asdata) in topology.bgp.as_list.items():
      if 'name' in asdata and asn in maps.bgp:
        maps.bgp[asn].name = asdata.name

  return maps

def graph_topology(topology: Box, fname: str, settings: Box,g_format: typing.Optional[list]) -> bool:
  f = common.open_output_file(fname)
  graph_start(f)

  maps = build_maps(topology)

  if 'bgp' in maps and settings.as_clusters:
    graph_bgp_clusters(f,maps.bgp,settings)
  else:
    for name,n in topology.nodes.items():
      node_with_label(f,n,settings)

  for l in topology.links:
    if l.type == "p2p":
      edge_p2p(f,l,settings.interface_labels)
    else:
      if not l.bridge:
        common.error('Found a lan/stub link without a bridge name, skipping',common.IncorrectValue,'graph')
        next
      network_with_label(f,l,settings)
      for ifdata in l.interfaces:
        if ifdata.node in maps.nodes:
          edge_node_net(f,l,ifdata.node,settings.interface_labels)

  f.write("}\n")
  f.close()
  return True

def bgp_session(f : typing.TextIO, node: Box, session: Box, settings: Box, rr_session: bool) -> None:
  arrow_dir = 'both'
  if rr_session:
    arrow_dir = 'none'
    if session.type == 'ibgp':
      if 'rr' in node.bgp and node.bgp.rr and not 'rr' in session:
        arrow_dir = 'forward'
      if not 'rr' in node.bgp and 'rr' in session:
        arrow_dir = 'back'

  f.write('  "%s" -- "%s"' % (node.name,session.name))
  f.write('  [\n')
  if session.type == 'ibgp':
    f.write('    color="%s"\n' % settings.colors.get('ibgp','#613913'))
  else:
    f.write('    color="%s"\n' % settings.colors.get('ebgp','#b21a1a'))
  f.write(f'    penwidth=2.5 arrowsize=0.7 dir={arrow_dir}\n')
  f.write('  ]\n')

def graph_bgp(topology: Box, fname: str, settings: Box,g_format: typing.Optional[list]) -> bool:
  if not 'bgp' in topology.get('module',{}):
    common.error('BGP graph format can only be used to draw topologies using BGP')
    return False

  f = common.open_output_file(fname)
  graph_start(f)

  rr_session = g_format is not None and len(g_format) > 1 and g_format[1] == 'rr'

  maps = build_maps(topology)
  graph_bgp_clusters(f,maps.bgp,settings)

  for name,n in topology.nodes.items():
    if 'bgp' in n:
      for neighbor in n.bgp.get('neighbors',[]):
        if neighbor.name > n.name:
          bgp_session(f,n,neighbor,settings,rr_session)

  f.write("}\n")
  f.close()
  return True

graph_dispatch = {
  'topology': graph_topology,
  'bgp': graph_bgp
}

class Graph(_TopologyOutput):

  def write(self, topology: Box) -> None:
    graphfile = self.settings.filename or 'graph.dot'
    output_format = 'topology'

    if hasattr(self,'filenames'):
      graphfile = self.filenames[0]
      if len(self.filenames) > 1:
        common.error('Extra output filename(s) ignored: %s' % str(self.filenames[1:]),common.IncorrectValue,'graph')

    if self.format:
      output_format = self.format[0]

    if output_format in graph_dispatch:
      if graph_dispatch[output_format](topology,graphfile,self.settings,self.format):
        print("Created graph file %s in %s format" % (graphfile, output_format))
    else:
      formats = ', '.join(graph_dispatch.keys())
      common.error('Unknown graph format, use one of %s' % formats,common.IncorrectValue,'graph')



#
# netlab read command
#
# Read network topology, add default settings, and dump the results
#
import typing
import sys
import argparse

from . import common_parse_args, topology_parse_args
from .. import read_topology,common,augment
from ..augment.main import transform_setup

#
# CLI parser for 'netlab read' command
#
def read_topology_parse(args: typing.List[str]) -> argparse.Namespace:
  parser = argparse.ArgumentParser(
    parents=[ common_parse_args(),topology_parse_args() ],
    prog="netlab read",
    description='Read network topology, add default settings, and dump the results')

  parser.add_argument(dest='topology', action='store', help='Topology file')
  parser.add_argument(
    '-o','--output',
    dest='output',
    type=argparse.FileType('w'),
    default=sys.stdout,
    action='store',
    help='Output file')
  return parser.parse_args(args)

def run(cli_args: typing.List[str]) -> None:
  args = read_topology_parse(cli_args)
  common.set_logging_flags(args)
  topology = read_topology.load(args.topology,args.defaults,"package:topology-defaults.yml")

  if 'settings' in args:
    topology.nodes = augment.nodes.create_node_dict(topology.nodes)
    read_topology.add_cli_args(topology,args)

  common.exit_on_error()

  transform_setup(topology)
  args.output.write(topology.to_yaml())



#
# netlab config command
#
# Deploy custom configuration template to network devices
#
import typing
import argparse
import os
import string
import pathlib
import glob
import subprocess
import shutil

from box import Box

from .. import common
from .. import read_topology
from . import external_commands

def package_parse(args: typing.List[str], settings: Box) -> argparse.Namespace:
  devs = [ k for k in settings.devices.keys() if settings.devices[k].libvirt.create or settings.devices[k].libvirt.create_template ]
  parser = argparse.ArgumentParser(
    prog='netlab libvirt package',
    description='Package a virtual machine into a libvirt Vagrant box')
  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='count',
    default=0,
    help='Verbose logging')
  parser.add_argument(
    '--skip',
    dest='skip',
    action='store',
    default='',help=argparse.SUPPRESS)
  parser.add_argument(
    dest='device',
    action='store',
    choices=devs,
    help='Network device you want to create')
  parser.add_argument(
    dest='disk',
    action='store',
    type=argparse.FileType('r'),
    help='Virtual machine disk (vmdk or qcow2)')
  return parser.parse_args(args)

def abort_on_failure(cmd: str) -> None:
  if not external_commands.run_command(cmd):
    common.fatal('Aborting')

def lp_create_vm_disk(args: argparse.Namespace) -> None:
  name = args.disk.name

  if 'vmdk' in name:
    print(f"Converting {name} into qcow2 format")
    abort_on_failure(f'qemu-img convert -f vmdk -O qcow2 {name} vm.qcow2')
  else:
    print(f"Creating a copy of {name}")
    abort_on_failure(f'cp {name} vm.qcow2')

def get_template_data(devdata: Box) -> Box:
  return devdata + { 'user' : { 'cwd' : os.getcwd() }}

def vm_cleanup(name: str) -> None:
  try:
    subprocess.run(f"virsh destroy {name}".split(" "))
  except Exception as ex:
    print(f"Cannot destroy {name}: {ex}")

  try:
    subprocess.run(f"virsh undefine {name}".split(" "))
  except Exception as ex:
    print(f"Cannot undefine {name}: {ex}")

def lp_create_bootstrap_iso(args: argparse.Namespace,settings: Box) -> None:
  devdata = settings.devices[args.device]
  if not 'create_iso' in devdata.libvirt:
    return
  print("Creating bootstrap ISO image")

  isodir = common.get_moddir() / "install/libvirt" / devdata.libvirt.create_iso
  shutil.rmtree('iso',ignore_errors=True)
  shutil.copytree(isodir,'iso')
  if os.path.exists('bootstrap.iso'):
    os.remove('bootstrap.iso')

  abort_on_failure("mkisofs -l -o bootstrap.iso iso")
  print("... done\n")

def lp_create_vm(args: argparse.Namespace,settings: Box) -> None:
  devdata = settings.devices[args.device]
  print(f"""
====================
Starting the VM
====================
We'll start the VM from the newly-created virtual disk. When the
VM starts, execute 'netlab libvirt config {args.device}' in another
window and follow the instructions.
====================

""")
  if devdata.libvirt.create:
    cmd = devdata.libvirt.create.replace("\n","")
    abort_on_failure(cmd)
  elif devdata.libvirt.create_template:
    data = get_template_data(devdata)
    template = common.template(devdata.libvirt.create_template,data,"install/libvirt",)
    pathlib.Path("template.xml").write_text(template)
    abort_on_failure("virsh define template.xml")
    abort_on_failure("virsh start --console vm_box")

  vm_cleanup('vm_box')

def lp_create_box(args: argparse.Namespace,settings: Box) -> None:
  with open("metadata.json","w") as metadata:
    metadata.write('{"provider":"libvirt","format":"qcow2","virtual_size":4}\n')
    metadata.close()
  print("Downloading vagrant-libvirt create_box.sh script")
  abort_on_failure('curl -O https://raw.githubusercontent.com/vagrant-libvirt/vagrant-libvirt/master/tools/create_box.sh')

  boxfile = f"{args.device}.box"
  if os.path.isfile(boxfile):
    print(f"Removing old {boxfile}")
    os.remove(boxfile)

  abort_on_failure(f'bash create_box.sh vm.qcow2 {boxfile}')

  devdata = settings.devices[args.device]
  boxname = devdata.libvirt.image
  if not boxname:
    common.fatal("Libvirt box name is not set for device {args.device}")

  print(f"""

=========================
Importing the Vagrant box
=========================
Your Vagrant box is ready to be imported. We just need a few
bits of information to tag it properly so you can have multiple
Vagrant boxes (different software versions) for the same network
device.

""")
  if not boxname:
    boxname = input('Enter box name: ')

  print(f"""
Your boxes should have versions. A box version can be anything; it's
best to use the version of the network operating system so you'll know
what your boxes do and be able to select a particular OS version in your
lab topology if you feel like building multiple boxes for the same OS.

You might want to limit yourself to using alphanumeric characters and dots.

Examples: 9.3.8 for Nexus OS, 4.27.0M for Arista EOS, 17.03.04 for CSR...

""")
  version = input('Enter box version: ')
  description = devdata.libvirt.description or devdata.description or (args.device+" box")
  json = string.Template("""
{
  "name": "$boxname",
  "description": "$description",
  "versions": [
    {
      "version": "$version",
      "providers": [
        {
          "name": "libvirt",
          "url": "file://$path/$device.box"
        }
      ]
    }
  ]
}
""")
  print("Creating box metadata in box.json")
  with open("box.json","w") as metadata:
    metadata.write(json.substitute(boxname=boxname,description=description,version=version,device=args.device,path=os.getcwd()))
    metadata.close()
  print("""

Importing Vagrant box
=====================
""")
  if not external_commands.run_command("vagrant box add box.json"):
    print("""

Failed to add Vagrant box. Fix the error(s) and use "vagrant box add box.json" to add it.
""")

def libvirt_package(cli_args: typing.List[str], settings: Box) -> None:
  args = package_parse(cli_args,settings)
  skip = args.skip

  print("""
=================
     WARNING
=================
This is an experimental script that does its best to build a Vagrant box for libvirt
provider out of a VM disk. It might die a horrible death and leave all sorts of garbage
behind that you'll have to clean up by hand (for example, libvirt 'vm_box' virtual machine).

It also assumes that it can wreak havoc in the current directory (although it will do its
best not to damage the original virtual disk).

""")
  if input('Do you want to continue [Y/n]: ') != 'Y':
    common.fatal('Aborting...')

  vm_cleanup('vm_box')
  if not 'disk' in skip:
    lp_create_vm_disk(args)
  if not 'iso' in skip:
    lp_create_bootstrap_iso(args,settings)
  if not 'vm' in skip:
    lp_create_vm(args,settings)
  if not 'box' in skip:
    lp_create_box(args,settings)

def config_parse(args: typing.List[str], settings: Box) -> argparse.Namespace:
  moddir = common.get_moddir()
  devs = map(
    lambda x: pathlib.Path(x).stem,
    glob.glob(str(moddir / "install/libvirt/*txt")))
  parser = argparse.ArgumentParser(
    prog='netlab libvirt config',
    description='Display Vagrant network device box configuration guidelines')
  parser.add_argument(
    dest='device',
    action='store',
    choices=list(devs),
    help='Network device you want to create')
  return parser.parse_args(args)

def libvirt_config(cli_args: typing.List[str], settings: Box) -> None:
  args = config_parse(cli_args,settings)
  helpfile = common.get_moddir() / "install/libvirt" / (args.device+".txt")
  print(helpfile.read_text())

def libvirt_usage() -> None:
  print("Usage: netlab libvirt package|config --help")

def run(cli_args: typing.List[str]) -> None:
  settings = read_topology.read_yaml('package:topology-defaults.yml')
  if not cli_args:
    libvirt_usage()
    return

  if not settings:
    common.fatal("Cannot read the system defaults","libvirt")
    return

  if cli_args[0] == 'package':
    libvirt_package(cli_args[1:],settings)
  elif cli_args[0] == 'config':
    libvirt_config(cli_args[1:],settings)
  else:
    libvirt_usage()



#
# VRF module
#
import typing
from box import Box

from . import _Module,_routing
from .. import common
from .. import data
from .. import addressing
from ..data import get_from_box
from ..augment import devices
from ..augment import links

# Global variables -- would love to be without them, but the alternatives
# are even messier
#
vlan_ids: Box
vlan_next: Box

# Static lists of keywords
#
vlan_mode_kwd: typing.Final[list] = [ 'bridge', 'irb', 'route' ]

vlan_link_attr: typing.Final[dict] = {
  'access': { 'type' : str, 'vlan': True, 'single': True },
  'native': { 'type' : str, 'vlan': True, 'single': True },
  'mode':   { 'type' : str },
  'trunk' : { 'type' : dict,'vlan': True }
}

phy_ifattr: typing.Final[list] = ['bridge','ifindex','parentindex','ifname','linkindex','type','vlan'] # Physical interface attributes
keep_subif_attr: typing.Final[list] = ['vlan','ifindex','ifname','type']    # Keep these attributes on VLAN subinterfaces

"""
init_global_vars: Initialize the VLAN ID pool
"""
def init_global_vars() -> None:
  global vlan_ids, vlan_next
  vlan_ids = Box({},default_box=True,box_dots=True)
  vlan_next = Box({},default_box=True,box_dots=True)

#
# build_vrf_id_set: given an object (topology or node), create a set of RDs
# that appear in that object.
#
def build_vlan_id_set(obj: Box, attr: str, objname: str) -> set:
  if 'vlans' in obj:
    if not isinstance(obj.vlans,dict):    # pragma: no cover
      common.fatal(f'Found a "vlans" setting that is not a dictionary in {objname}','vlan')
      return set()

    return { v[attr] for v in obj.vlans.values() if isinstance(v,dict) and attr in v and v[attr] is not None }
  return set()

def populate_vlan_id_set(topology: Box) -> None:
  global vlan_ids, vlan_next

  for k in ('id','vni'):
    vlan_ids[k] = build_vlan_id_set(topology,k,'topology')

  vlan_next['id'] = topology.defaults.vlan.start_vlan_id
  vlan_next['vni'] = topology.defaults.vlan.start_vni

  for n in topology.nodes.values():
    if 'vlans' in n:
      for k in ('id','vni'):
        vlan_ids[k] = vlan_ids[k].union(build_vlan_id_set(n,k,n.name))

def get_next_vlan_id(k : str) -> int:
  global vlan_ids,vlan_next

  if k not in vlan_ids or k not in vlan_next:   # pragma: no cover
    common.fatal(f'Invalid attribute {k} in get_next_vlan_id call')

  while vlan_next[k] in vlan_ids[k]:
    vlan_next[k] = vlan_next[k] + 1

  vlan_ids[k].add(vlan_next[k])
  return vlan_next[k]

#
# Validate VLAN attributes and set missing attributes:
#
# * VLAN and VNI
# * Subnet prefix
#
def validate_vlan_attributes(obj: Box, topology: Box) -> None:
  global vlan_ids

  obj_name = 'global VLANs' if obj is topology else obj.name
  default_fwd_mode = get_from_box(obj,'vlan.mode')                # Get node-wide VLAN forwarding mode
  if default_fwd_mode:                                            # ... is it set?
    default_fwd_mode = str(default_fwd_mode)                      # Convert it to string so we don't have to deal with weird types
    if not default_fwd_mode in vlan_mode_kwd:                     # ... and check keyword validity
      common.error(
        f'Invalid vlan.mode setting {default_fwd_mode} in {obj_name}',
        common.IncorrectValue,
        'vlan')

  if not 'vlans' in obj:
    return

  for vname in list(obj.vlans.keys()):
    if not obj.vlans[vname]:
      obj.vlans[vname] = Box({},default_box=True,box_dots=True)

    vdata = obj.vlans[vname]

    if 'mode' in vdata:                                             # Do we have 'mode' set in the VLAN definition?
      if not vdata.mode in vlan_mode_kwd:                           # ... check the keyword value 
        common.error(
          f'Invalid VLAN mode setting {vdata.mode} in VLAN {vname} in {obj_name}',
          common.IncorrectValue,
          'vlan')
    else:
      if default_fwd_mode:                                          # Propagate default VLAN forwarding mode if needed
        vdata.mode = default_fwd_mode

    if 'mode' in vdata and vdata.mode == 'route':                   # Throw out attempts to have routed interfaces
      common.error(f'VLAN routed interfaces not yet supported: VLAN {vname} in {obj_name}',common.IncorrectValue,'vlan')
    if not 'id' in vdata:                                           # When VLAN ID is not defined
      vdata.id = get_next_vlan_id('id')                             # ... take the next free VLAN ID from the list
    if not isinstance(vdata.id,int):                                # Now validate the heck out of VLAN ID
      common.error(f'VLAN ID {vdata.id} for VLAN {vname} in {obj_name} must be an integer',common.IncorrectValue,'vlan')
      continue
    if vdata.id < 2 or vdata.id > 4094:
      common.error(f'VLAN ID {vdata.id} for VLAN {vname} in {obj_name} must be between 2 and 4094',common.IncorrectValue,'vlan')
      continue

    if not 'vni' in vdata:                                          # When VNI is not defined
      vni_default = topology.defaults.vlan.start_vni + vdata.id     # ... try to build VNI from VLAN ID
      if not vni_default in vlan_ids.vni:                           # Is the VNI free?
        vdata.vni = vni_default                                     # ... great, take it
        vlan_ids.vni.add(vni_default)                               # ... and add it to the list of used VNIs
      else:                                                         # Too bad, we had such a great idea but it failed
        vdata.vni = get_next_vlan_id('vni')                         # ... so take the next available VNI
    if not isinstance(vdata.vni,int):                               # Not done yet, we still have to validate the VNI type and range
      common.error(f'VNI {vdata.vni} for VLAN {vname} in {obj_name} must be an integer',common.IncorrectValue,'vlan')
      continue
    if vdata.vni < 2 or vdata.vni > 16777215:
      common.error(f'VNI {vdata.vni} for VLAN {vname} in {obj_name} must be between 2 and 16777215',common.IncorrectValue,'vlan')
      continue

    vlan_pool = [ vdata.pool ] if 'pool' in vdata else []
    vlan_pool.extend(['vlan','lan'])
    pfx_list = links.augment_link_prefix(vdata,vlan_pool,topology.pools)
    vdata.prefix = addressing.rebuild_prefix(pfx_list)

"""
check_link_vlan_attributes: check correctness of VLAN link attributes

* known attributes only
* attribute types must be correct
* VLANs used in attribute types must be defined
"""

def check_link_vlan_attributes(obj: Box, link: Box, v_attr: Box, topology: Box) -> bool:
  global vlan_link_attr
  if not 'vlan' in obj:
    return True

  if not isinstance(obj.vlan,dict):                               # Basic sanity check: VLAN attribute must be a dictionary
    common.error(
      f'vlan link attribute must be a dictionary\n... {link}',
      common.IncorrectValue,
      'vlan')
    return False

  node_error = f' in node {obj.node}' if not obj is link else ''  # Prepare for error checking
  link_ok = True

  for attr in obj.vlan.keys():                                    # Check for unexpected attributes
    if not attr in vlan_link_attr:
      common.error(
        f'Unknown VLAN attribute {attr}{node_error}\n... {link}',
        common.IncorrectValue,
        'vlan')
      link_ok = False

  for attr in vlan_link_attr.keys():                              # Loop over VLAN attributes
    if not attr in obj.vlan:                                      # ... not present, skip
      continue

    a_type = vlan_link_attr[attr]['type']
    if not isinstance(obj.vlan[attr],a_type):                     # Is the attribute type correct?
      if a_type is dict and isinstance(obj.vlan[attr],list):      # ... only exception: convert list to dict
        obj.vlan[attr] = { vname: {} for vname in obj.vlan[attr] }
      else:
        common.error(
          f'VLAN attribute {attr}{node_error} must be a {a_type.__name__}\n... {link}',
          common.IncorrectValue,
          'vlan')  
        link_ok = False
        continue

    if isinstance(obj.vlan[attr],dict):                           # Change empty VLAN keys into dictionaries
      for k in obj.vlan[attr].keys():
        if obj.vlan[attr][k] is None:
          obj.vlan[attr][k] = {}

    if not 'vlan' in vlan_link_attr[attr]:                        # If this attribute does not contain VLAN names, skip the rest
      continue
                                                                  # Build a list of VLANs out of a string or a dict
    vlan_list = [ obj.vlan[attr] ] if isinstance(obj.vlan[attr],str) else obj.vlan[attr].keys()
    if not attr in v_attr:                                        # Prepare attribute collection dictionary if needed
      v_attr[attr].list = []
      v_attr[attr].set = set()
      v_attr[attr].node_set = set()
      v_attr[attr].use_count = 0

    v_attr[attr].list.extend(vlan_list)                           # ... and add collected VLANs to attribute dictionary
    v_attr[attr].set.update(vlan_list)
    v_attr[attr].use_count = v_attr[attr].use_count + 1

    for vname in vlan_list:                                       # Check that VLANs exist
      if vname in topology.get('vlans',{}):
        continue
      if not obj is link and vname in topology.nodes[obj.node].get('vlans',{}):
        v_attr[attr].node_set.add(obj.node)
        continue
      common.error(
        f'VLAN {vname} used in vlan.{attr}{node_error} is not defined\n... {link}',
        common.IncorrectValue,
        'vlan')
      link_ok = False

  return link_ok

"""
validate_link_vlan_attributes: check semantical correctness of VLAN attributes

* 'trunk' and 'access' cannot be mixed
* 'native' is valid only with 'trunk'
* 'access' and 'native' should have a single value
"""
def validate_link_vlan_attributes(link: Box,v_attr: Box,topology: Box) -> bool:
  global vlan_link_attr

  if 'trunk' in v_attr and 'access' in v_attr:
    common.error(
      f"Cannot mix trunk and access VLANs on the same link\n... {link}",
      common.IncorrectValue,
      'vlan')
    return False

  if 'native' in v_attr and not 'trunk' in v_attr:
    common.error(
      f"Native VLAN is valid only on VLAN trunks\n... {link}",
      common.IncorrectValue,
      'vlan')
    return False

  link_ok = True

  for attr in vlan_link_attr.keys():                              # Loop over VLAN attributes
    if not attr in v_attr:                                        # ... not present, skip
      continue

    if not 'vlan' in v_attr:                                      # Not a list of VLANs, no further checks necessary
      continue

    if v_attr[attr].use_count > 1:                                # Is this attribute used in more than one place on the link?
      for vname in v_attr[attr].set:                              # Iterate over the VLAN set
        if not vname in topology.get('vlans',{}):                 # ... and check that all VLANs used this way are globally defined
          common.error(
            f"VLAN {vname} used in more than one place on the same link must be a global VLAN\n... {link}",
            common.IncorrectValue,
            'vlan')
          link_ok = False

    if not 'single' in v_attr:                                    # Does the attribute require a consistent VLAN across the link?
      continue

    if len(v_attr[attr].set) > 1:                                 # ... if so, check the length of the VLAN set
      common.error(
        f"Cannot use more than one {attr} VLAN on the same link\n... {link}",
        common.IncorrectValue,
        'vlan')
      return False

  return link_ok

"""
copy_vlan_attributes: copy prefix and link type from vlan to link
"""
def copy_vlan_attributes(vlan: str, vlan_data: Box, link: Box) -> None:
  if 'prefix' in vlan_data:
    link.prefix = vlan_data.prefix
  if not 'type' in link:
    link.type = vlan_data.get('type','lan')
  link.vlan_name = vlan

"""
set_link_vlan_prefix: copy link attributes from VLAN for access/native VLAN links

* We're assuming the attributes passed VLAN validity checks, so it's safe to
  use the collected v_attr data
* If there's access or native VLAN defined on the link, use it to set link attributes
"""

def set_link_vlan_prefix(link: Box, v_attr: Box, topology: Box) -> None:
  link_vlan_set: set = set()
  node_set: set = set()

  if 'access' in v_attr:
    link_vlan_set = v_attr.access.set
    node_set = v_attr.access.node_set
  elif 'native' in v_attr:
    link_vlan_set = v_attr.native.set
    node_set = v_attr.native.node_set
  elif 'trunk' in v_attr:
    if not 'role' in link and not 'prefix' in link:   # If the user set prefix or address pool leave it alone
      link.prefix = {}                                # ... otherwise we need no IP addressing on trunk links without a native VLAN
    return

  if not link_vlan_set:
    return

  link_vlan = list(link_vlan_set)[0]                  # Got the access/native VLAN
  if link_vlan in topology.get('vlans',{}):
    copy_vlan_attributes(link_vlan,topology.vlans[link_vlan],link)
    return

  if not node_set:
    common.fatal(f'Cannot find the node using VLAN {link_vlan}\n... {link}')
    return

  copy_vlan_attributes(link_vlan,topology.nodes[list(node_set)[0]].vlans[link_vlan],link)

"""
create_vlan_links: Create virtual links for every VLAN in the VLAN trunk
"""
def create_vlan_links(link: Box, v_attr: Box, topology: Box) -> None:
  native_vlan = link.get('vlan_name',None)
  for vname in sorted(v_attr.trunk.set):
    if vname != native_vlan:           # Skip native VLAN
      link_data = Box(link.vlan.trunk[vname] or {},default_box=True,box_dots=True)
      link_data.linkindex = topology.links[-1].linkindex + 1
      link_data.parentindex = link.linkindex
      link_data.vlan.access = vname
      link_data.vlan_name = vname
      link_data.type = 'vlan_member'
      link_data.interfaces = []
      for intf in link.interfaces:
        if 'vlan' in intf and vname in intf.vlan.get('trunk',{}):
          intf_data = Box(intf.vlan.trunk[vname] or {},default_box=True,box_dots=True)
          intf_data.node = intf.node
          link_data.interfaces.append(intf_data)
      topology.links.append(link_data)

"""
get_vlan_data: Get VLAN data structure (node or topology)
"""
def get_vlan_data(vlan: str, node: Box, topology: Box) -> Box:
  return topology.vlans[vlan] if vlan in topology.get('vlans',{}) else node.vlans[vlan]

"""
update_vlan_neighbor_list: Build a VLAN-wide list of neighbors
"""
def update_vlan_neighbor_list(vlan: str, phy_if: Box, svi_if: Box, node: Box,topology: Box) -> None:
  vlan_data = get_vlan_data(vlan,node,topology)
  if not 'neighbors' in vlan_data:
    vlan_data.neighbors = []

  n_map = { data.node: data for data in vlan_data.neighbors }           # Build a list of known VLAN neighbors
  phy_n_list = phy_if.get('neighbors',[])                               # ... add interface neighbors not yet in the list
  vlan_data.neighbors.extend([n_data for n_data in phy_n_list if n_data.node not in n_map])

  if node.name in n_map:                                                # Is the current node in the list?
    n_map[node.name].ifname = svi_if.ifname                             # ... it is, fix the interface name
    for af in ('ipv4','ipv6'):
      if af in svi_if:
        n_map[node.name][af] = svi_if[af]
      else:
        n_map[node.name].pop(af,None)
  else:
    n_data = { 'ifname': svi_if.ifname, 'node': node.name }             # ... not yet, create neighbor data
    for af in ('ipv4','ipv6'):
      if af in svi_if:                                                  # ... copy SVI interface addresses to neighbor data
        n_data[af] = svi_if[af]
    vlan_data.neighbors.append(n_data)                                  # Add current node as a neighbor to VLAN neighbor list

"""
create_node_vlan: Create a local (node) copy of a VLAN used on an interface
"""
def create_node_vlan(node: Box, vlan: str, topology: Box) -> typing.Optional[Box]:
  if not vlan in node.vlans:                                        # Do we have VLAN defined in the node?
    node.vlans[vlan] = Box(topology.vlans[vlan])                    # ... no, create a copy of the global definition
    if not node.vlans[vlan]:                                        # pragma: no cover -- we don't have a global definition? 
      common.fatal(                                                 # ... this should have been detected way earlier
        f'Unknown VLAN {vlan} used on node {node.name}','vlan')
      return None

  if not 'bridge_group' in node.vlans[vlan]:                        # Set bridge group in VLAN data
    if not node.vlan.max_bridge_group:                              # ... current counter is in node.vlan dictionary
      node.vlan.max_bridge_group = 1
    else:
      node.vlan.max_bridge_group = node.vlan.max_bridge_group + 1
    node.vlans[vlan].bridge_group = node.vlan.max_bridge_group

  return node.vlans[vlan]

"""
create_svi_interfaces: for every physical interface with access VLAN, create an SVI interface
"""

def create_svi_interfaces(node: Box, topology: Box) -> dict:
  global phy_ifattr
  vlan_ifmap: dict = {}

  svi_skipattr = ['id','vni','prefix','pool']                               # VLAN attributes not copied into VLAN interface
  iflist_len = len(node.interfaces)
  for ifidx in range(0,iflist_len):
    ifdata = node.interfaces[ifidx]
    access_vlan = get_from_box(ifdata,'vlan.access') or get_from_box(ifdata,'vlan.native')
    if not access_vlan:                                                     # No access VLAN on this interface?
      continue                                                              # ... good, move on

    vlan_data = create_node_vlan(node,access_vlan,topology)
    if vlan_data is None:
      continue

    features = devices.get_device_features(node,topology.defaults)
    svi_name = features.vlan.svi_interface_name
    if not svi_name:                                                        # pragma: no cover -- hope we got device settings right ;)
      common.error(                                                         # SVI interfaces are not supported on this device
        f'Device {node.device} used by {node.name} does not support VLAN interfaces (access vlan {access_vlan})',
        common.IncorrectValue,
        'vlan')
      return vlan_ifmap

    if not access_vlan in vlan_ifmap:                                       # Do we need to create a SVI interface?
      skip_attr = list(phy_ifattr)                                          # Create a local copy of the attribute skip list
      if vlan_data.get('mode','') == 'bridge':                              # ... and skip IP addresses for bridging-only VLANs
        skip_attr.extend(['ipv4','ipv6'])
      vlan_ifdata = Box(                                                    # Copy non-physical interface attributes into SVI interface
        { k:v for k,v in ifdata.items() if k not in skip_attr },            # ... that will also give us IP addresses
        default_box=True,
        box_dots=True)
      vlan_ifdata.ifindex = node.interfaces[-1].ifindex + 1                 # Fill in the rest of interface data:
      vlan_ifdata.ifname = svi_name.format(                                 # ... ifindex, ifname, description
                              vlan=vlan_data.id,
                              bvi=vlan_data.bridge_group)
      vlan_ifdata.name = f'VLAN {access_vlan} ({vlan_data.id})'
      vlan_ifdata.virtual_interface = True                                  # Mark interface as virtual
      vlan_ifdata.type = "svi"
      vlan_ifdata.neighbors = []                                            # No neighbors so far
                                                                            # Overwrite interface settings with VLAN settings
      vlan_ifdata = vlan_ifdata + { k:v for k,v in vlan_data.items() if k not in svi_skipattr }
      node.interfaces.append(vlan_ifdata)                                   # ... and add SVI interface to list of node interfaces
      vlan_ifmap[access_vlan] = vlan_ifdata
    else:
      vlan_ifdata = vlan_ifmap[access_vlan]

    update_vlan_neighbor_list(access_vlan,ifdata,vlan_ifdata,node,topology)

    for attr in list(ifdata.keys()):                                        # Clean up physical interface data
      if not attr in phy_ifattr:
        ifdata.pop(attr,None)

    ifdata.vlan.access_id = vlan_data.id                                    # Add VLAN ID to interface data to simplify config templates

  return vlan_ifmap

"""
set_svi_neighbor_list: set SVI neighbor list from VLAN neighbor list
"""
def set_svi_neighbor_list(node: Box, topology: Box) -> None:
  for ifdata in node.interfaces:
    if 'vlan_name' in ifdata:
      vlan_data = get_vlan_data(ifdata.vlan_name,node,topology)
      if not 'host_count' in vlan_data:                                     # Calculate the number of hosts attached to the VLAN
        vlan_data.host_count = len([ n for n in vlan_data.neighbors if topology.nodes[n.node].get('role','') == 'host' ])

      if not 'role' in ifdata:                                              # If the interface has no 'role' copied from the link
        if len(vlan_data.neighbors) == vlan_data.host_count +1:             # ... and there's exactly one non-host attached to the VLAN
          ifdata.role = 'stub'                                              # ... then we have a stub link, mark it for IGP modules

      ifdata.neighbors = [ n for n in vlan_data.neighbors if n.node != node.name ]
      if ifdata.neighbors:
        if ' -> ' in ifdata.name:
          ifdata.name = ifdata.name.split(' -> ')[0]
        ifdata.name = ifdata.name + " -> [" + ",".join([ n.node for n in ifdata.neighbors]) + "]"

"""
map_trunk_vlans: build a list of VLAN IDs on trunk interfaces
"""
def map_trunk_vlans(node: Box, topology: Box) -> None:
  for intf in node.interfaces:
    trunk = get_from_box(intf,'vlan.trunk')
    if not trunk:
      continue

    vlan_list = []
    for vlan in trunk.keys():
      if not vlan in node.vlans:
        common.fatal(f'Internal error: VLAN {vlan} should be already defined on node {node.name}')
        break
      vlan_list.append(node.vlans[vlan].id)

    intf.vlan.trunk_id = vlan_list

"""
find_parent_interface: Find the parent interface of a VLAN member subinterface
"""
def find_parent_interface(intf: Box, node: Box, topology: Box) -> typing.Optional[Box]:
  link_list = [ l for l in topology.links if l.linkindex == intf.parentindex ]
  if not link_list:
    return None

  link = link_list[0]
  intf_list = [ intf for intf in link.interfaces if intf.node == node.name]
  if not intf_list:
    return None

  link_intf = intf_list[0]

  node_iflist = [ intf for intf in node.interfaces if intf.ifindex == link_intf.ifindex]
  if not node_iflist:
    return None

  return node_iflist[0]

"""
rename_vlan_subinterfaces: rename or remove interfaces created from VLAN pseudo-links
"""
def rename_vlan_subinterfaces(node: Box, topology: Box) -> None:
  global phy_ifattr

  features = devices.get_device_features(node,topology.defaults)
  subif_name = features.vlan.vlan_subif_name

  if not subif_name:
    node.interfaces = [ intf for intf in node.interfaces if intf.type != 'vlan_member']
    return

  for intf in node.interfaces:
    if intf.type != 'vlan_member':
      continue

    parent_intf = find_parent_interface(intf,node,topology)
    if parent_intf is None:
      common.fatal(f'Internal error: cannot find parent interface for {intf} in node {node.name}')
      return

    if 'subif_index' in parent_intf:
      parent_intf.subif_index = parent_intf.subif_index + 1
    else:
      parent_intf.subif_index = 1

    intf.ifname = subif_name.format(**parent_intf)
    intf.parent_ifindex = parent_intf.ifindex
    intf.virtual_interface = True
    for attr in phy_ifattr:
      if attr in intf and not attr in keep_subif_attr:
        intf.pop(attr,None)

class VLAN(_Module):

  def module_pre_transform(self, topology: Box) -> None:
    init_global_vars()
    if get_from_box(topology,'vlan.mode'):
      if topology.vlan.mode not in vlan_mode_kwd:     # pragma: no cover
        common.error(
          f'Invalid global vlan.mode value {topology.vlan.mode}',
          common.IncorrectValue,
          'vlan')
        return

    populate_vlan_id_set(topology)

    if not 'vlans' in topology:
      return

    validate_vlan_attributes(topology,topology)

  def node_pre_transform(self, node: Box, topology: Box) -> None:
    if 'vlans' in node:
      for vname in node.vlans.keys():
        if 'vlans' in topology and vname in topology.vlans:
          for kw in ('prefix','id','vni'):
            if kw in node.vlans[vname]:
              common.error(
                f'Cannot set {kw} for VLAN {vname} on node {node.name} -- VLAN is defined globally',
                common.IncorrectValue,
                'vlan')
          node.vlans[vname] = topology.vlans[vname] + node.vlans[vname]

    validate_vlan_attributes(node,topology)

  def link_pre_transform(self, link: Box, topology: Box) -> None:
    if link.get('type','') == 'vlan_member':                                      # Skip VLAN member links, we've been there...
      return

    v_attr = Box({},default_box=True,box_dots=True)
    link_ok = check_link_vlan_attributes(link,link,v_attr,topology)               # Check link-level VLAN attributes

    for intf in link.interfaces:
      link_ok = link_ok and check_link_vlan_attributes(intf,link,v_attr,topology) # Check interface VLAN attributes

    if not link_ok:
      return

    if not validate_link_vlan_attributes(link,v_attr,topology):
      return

    set_link_vlan_prefix(link,v_attr,topology)

    # Merge link VLAN attributes into interface VLAN attributes to make subsequent steps easier
    if 'vlan' in link:
      for intf in link.interfaces:
        if 'vlan' in topology.nodes[intf.node].get('module',[]):
          intf.vlan = link.vlan + intf.vlan

    if 'trunk' in v_attr:
      create_vlan_links(link,v_attr,topology)

  def module_post_transform(self, topology: Box) -> None:
    for n in topology.nodes.values():
      if 'vlan' in n.get('module',[]):
        vlan_ifmap = create_svi_interfaces(n,topology)
        map_trunk_vlans(n,topology)
        rename_vlan_subinterfaces(n,topology)

    for n in topology.nodes.values():
      set_svi_neighbor_list(n,topology)

    topology.links = [ link for link in topology.links if link.type != 'vlan_member' ]



#
# Common Ansible interface commands
#

import typing
import subprocess
import sys
import os
import json
from pathlib import Path

from .. import common

try:
  from importlib import resources
except ImportError:
  import importlib_resources as resources # type: ignore

def find_playbook(name: str) -> typing.Union[str,None]:
  cwd = Path(os.getcwd()).resolve()
  scriptdir = Path(sys.argv[0]).resolve().parent
  moddir = Path(__file__).resolve().parent.parent

  for dir in [cwd,cwd / 'ansible',scriptdir / 'ansible',moddir / 'ansible']:
    if os.path.isfile(dir / name):
      return str(dir / name)

  return None

def inventory(name: str) -> typing.Optional[dict]:
  try:
    result = subprocess.run(['ansible-inventory','--host',name],capture_output=True,check=True,text=True)
    try:
      return json.loads(result.stdout)
    except:
      common.fatal('Cannot parse JSON data returned by ansible-inventory','inventory')
    return None

  except:
    try:
      subprocess.run(['ansible-inventory','-h'],capture_output=True,check=True)
    except Exception as ex:
      common.fatal(f'Cannot execute ansible-inventory command\n  {ex}','inventory')

    common.fatal('Cannot get Ansible inventory data for %s with ansible-inventory. Is the host name correct?' % name,'inventory')

  return None

def playbook(name: str, args: typing.List[str]) -> None:
  pbname = find_playbook(name)
  if not pbname:
    common.fatal("Cannot find Ansible playbook %s, aborting" % name)
    return

  if common.VERBOSE:
    print("Running Ansible playbook %s" % pbname)

  cmd = ['ansible-playbook',pbname]
  cmd.extend(args)

  try:
    subprocess.check_call(cmd)
  except Exception as ex:
    common.fatal(f"Executing Ansible playbook {pbname} failed:\n  {ex}")



#
# netlab connect command
#
# Connect to a lab device using SSH or Docker
#
import typing
import os
import sys
import argparse
import subprocess

from box import Box

from . import ansible
from ..outputs import devices
from . import common_parse_args
from .. import common

#
# CLI parser for 'netlab initial' command
#
def connect_parse(args: typing.List[str]) -> typing.Tuple[argparse.Namespace, typing.List[str]]:
  parser = argparse.ArgumentParser(
    prog="netlab connect",
    description='Connect to a network device',
    epilog='The rest of the arguments are passed to SSH or docker exec command')
  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='store_true',
    help='Verbose logging')
  parser.add_argument(
    '-d','--devices',
    dest='devices',
    action='store_true',
    help='Use netsim-devices.yml as inventory source')
  parser.add_argument(
    dest='host', action='store',
    help='Device to connect to')

  return parser.parse_known_args(args)

def get_inventory_data(host: str,source: str) -> typing.Optional[dict]:
  if source == 'ansible':
    return ansible.inventory(host)
  elif source == 'devices':
    return devices.read_inventory(host)
  else:
    common.fatal('Unknown inventory source %s' % source,'connect')
    return None

def docker_connect(data: Box, rest: typing.List[str], verbose: bool = False) -> None:
  host = data.ansible_host or data.host

  shell = data.get('docker_shell','bash')
  args = ['docker','exec','-it',host,shell,'-il']
  if rest:
    sys.stderr.write("Connecting to container %s, executing %s\n" % (host," ".join(rest)))
    args.extend(['-c','"'+' '.join(rest)+'"'])
  else:
    sys.stderr.write("Connecting to container %s, starting bash\n" % host)
  sys.stderr.flush()
  subprocess.run(args)

def ssh_connect(data: Box, rest: typing.List[str], verbose: bool = False) -> None:
  host = data.ansible_host or data.host
  args = ['ssh','-o','UserKnownHostsFile=/dev/null','-o','StrictHostKeyChecking=no','-o','LogLevel ERROR']

  if data.ansible_ssh_pass:
    args = ['sshpass','-p',data.ansible_ssh_pass ] + args

  if data.ansible_port:
    args.extend(['-p',str(data.ansible_port)])

  if data.ansible_user:
    args.extend([data.ansible_user+"@"+host])
  else:
    args.extend([host])

  args.extend(rest)
  if verbose:
    print("Executing: %s" % args)
  else:
    sys.stderr.write("Connecting to %s using SSH port %s\n" % (host,data.ansible_port or 22))
    sys.stderr.flush()

  subprocess.run(args)

def run(cli_args: typing.List[str]) -> None:
  (args,rest) = connect_parse(cli_args)
  inventory_source = 'devices' if args.devices else 'ansible'

  host_inventory = get_inventory_data(args.host,inventory_source)
  if not host_inventory:
    common.fatal('Hostname %s not found in %s inventory' % (args.host,inventory_source))
    return

  host_data = Box(host_inventory,box_dots=True,default_box=True)
  host_data.host = args.host
  connection = host_data.netsim_console_connection or host_data.ansible_connection

  if connection == 'docker':
    docker_connect(host_data,rest,args.verbose)
  elif connection in ['paramiko','ssh','network_cli','netconf','httpapi',] or not connection:
    if connection in ['netconf','httpapi']:
      print(f"Using SSH to connect to a device configured with {connection} connection")
    ssh_connect(host_data,rest,args.verbose)
  else:
    common.fatal('Unknown connection method %s for host %s' % (connection,args.host),'connect')



#
# netlab config command
#
# Deploy custom configuration template to network devices
#
import typing
import argparse
import os
import glob
import subprocess
import shutil

from box import Box
from pathlib import Path

from .. import common
from .. import read_topology
from . import external_commands

#
# CLI parser for 'netlab config' command
#
def test_parse(args: typing.List[str], settings: Box) -> argparse.Namespace:
  choices = ( p for p in settings.providers.keys() if 'start' in settings.providers[p] )
  parser = argparse.ArgumentParser(
    prog='netlab test',
    description='Test virtual lab installation')
  parser.add_argument(
    '-w','--work-directory',
    dest='workdir',
    action='store',
    default='test',
    help='Working directory (default: test)')
  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='count',
    default=0,
    help='Verbose logging')
  parser.add_argument(
    dest='provider',
    action='store',
    choices=list(choices),
    help='Run tests for the specified provider')

  return parser.parse_args(args)

def copy_topology(args: argparse.Namespace) -> None:
  if args.verbose:
    print("Creating test topology file in %s" % args.workdir)
  os.mkdir(args.workdir)
  os.chdir(args.workdir)
  moddir = Path(__file__).resolve().parent.parent

  src_topo = moddir / "templates/tests" / (args.provider + ".yml")
  try:
    shutil.copy(src_topo,"topology.yml")
  except:
    common.fatal("Cannot copy test topology file %s into %s" % (src_topo,args.workdir))
  if args.verbose:
    print("... done, moving on\n")

def create_configs() -> None:
  external_commands.print_step(2,"creating configuration files")
  if not external_commands.run_command(["netlab","create"],True):
    common.fatal("netlab create failed, aborting...","netlab test")

def cleanup_working_directory(args: argparse.Namespace) -> None:
  if args.verbose:
    print("\nCleaning up -- removing %s and its contents" % args.workdir)
  os.chdir("..")
  try:
    shutil.rmtree(args.workdir)
  except:
    print("... cannot clean up working directory, trying to run 'sudo rm -fr'")
    try:
      subprocess.run(["sudo","rm","-fr",args.workdir],check=True)
      print("... that worked ;)")
    except:
      print("... rm -fr failed, please clean up %s manually" % args.workdir)
      return

  if args.verbose:
    print("... done, test completed\n")

def run(cli_args: typing.List[str]) -> None:
  settings = read_topology.read_yaml('package:topology-defaults.yml')
  if not cli_args:
    common.fatal("Specify the virtualization environment to test or use -h to get help","test")
    return

  if not settings:
    common.fatal("Cannot read the global defaults","test")
    return

  args = test_parse(cli_args,settings)
  if os.path.exists(args.workdir):
    common.fatal("Directory %s already exists, aborting" % args.workdir,"test")

  if args.verbose:
    common.set_verbose(args.verbose)
  external_commands.run_probes(settings,args.provider,1)
  copy_topology(args)
  create_configs()
  if not external_commands.run_command('netlab up'):
    common.fatal('netlab up failed, aborting','test')
  elif not external_commands.run_command('netlab down'):
    common.fatal('netlab down failed','test')
#  external_commands.start_lab(settings,args.provider,3)
#  external_commands.deploy_configs(4)
#  external_commands.stop_lab(settings,args.provider,5)
  cleanup_working_directory(args)



#
# Containerlab provider module
#
import subprocess
import typing
from box import Box

from . import _Provider
from .. import common

def list_bridges( topology: Box ) -> typing.Set[str]:
    return { l.bridge for l in topology.links if l.bridge and l.node_count != 2 }

def use_ovs_bridge( topology: Box ) -> bool:
    return topology.defaults.providers.clab.bridge_type == "ovs-bridge"

def create_linux_bridge( brname: str ) -> bool:
    try:
      result = subprocess.run(['sudo','ip','link','add','name',brname,'type','bridge'],capture_output=True,check=True,text=True)
      common.print_verbose( f"Create Linux bridge '{brname}': {result}" )
      result2 = subprocess.run(['sudo','ip','link','set','dev',brname,'up'],capture_output=True,check=True,text=True)
      common.print_verbose( f"Enable Linux bridge '{brname}': {result2}" )
      result3 = subprocess.run(['sudo','sh','-c',f'echo 65528 >/sys/class/net/{brname}/bridge/group_fwd_mask'],check=True)
      common.print_verbose( f"Enable LLDP,LACP,802.1X forwarding on Linux bridge '{brname}': {result3}" )
      return True
    except Exception as ex:
      print(ex)
      common.error("Error creating bridge '%s': %s" % (brname,ex), module='clab')
      return False

def destroy_linux_bridge( brname: str ) -> bool:
    try:
      result = subprocess.run(['sudo','ip','link','del','dev',brname],capture_output=True,check=True,text=True)
      common.print_verbose( f"Delete Linux bridge '{brname}': {result}" )
      return True
    except Exception as ex:
      print(ex)
      common.error("Error deleting Linux bridge '%s': %s" % (brname,ex), module='clab')
      return False

def create_ovs_bridge( brname: str ) -> bool:
    try:
      result = subprocess.run(['sudo','ovs-vsctl','add-br',brname],capture_output=True,check=True,text=True)
      common.print_verbose( f"Create OVS bridge '{brname}': {result}" )
      return True
    except Exception as ex:
      print(ex)
      common.error("Error deleting OVS bridge '%s': %s" % (brname,ex), module='clab')
      return False

def destroy_ovs_bridge( brname: str ) -> bool:
    try:
      result = subprocess.run(['sudo','ovs-vsctl','del-br',brname],capture_output=True,check=True,text=True)
      common.print_verbose( f"Delete OVS bridge '{brname}': {result}" )
      return True
    except Exception as ex:
      print(ex)
      common.error("Error deleting OVS bridge '%s': %s" % (brname,ex), module='clab')
      return False

class Containerlab(_Provider):

  def augment_node_data(self, node: Box, topology: Box) -> None:
    node.hostname = "clab-%s-%s" % (topology.name,node.name)

  def pre_start_lab(self, topology: Box) -> None:
    common.print_verbose('pre-start hook for Containerlab - create any bridges')
    for brname in list_bridges(topology):
        if use_ovs_bridge(topology):
            create_ovs_bridge(brname)
        else:
            create_linux_bridge(brname)

  def post_stop_lab(self, topology: Box) -> None:
    common.print_verbose('post-stop hook for Containerlab, cleaning up any bridges')
    for brname in list_bridges(topology):
        if use_ovs_bridge(topology):
            destroy_ovs_bridge(brname)
        else:
            destroy_linux_bridge(brname)



#
# Read topology and default settings files
#
import os
import sys
import typing
import argparse

from box import Box
try:
  from importlib import resources
except ImportError:
  import importlib_resources as resources # type: ignore

# Related modules
from . import common
from . import data

#
# Read YAML from file, package file, or string
#
def read_yaml(filename: typing.Optional[str] = None, string: typing.Optional[str] = None) -> typing.Optional[Box]:
  if string is not None:
    try:
      yaml_data = Box().from_yaml(yaml_string=string,default_box=True,box_dots=True,default_box_none_transform=False)
    except:                                                                    # pragma: no cover -- can't get here unless there's a package error
      common.fatal("Cannot parse YAML string: %s " % (str(sys.exc_info()[1])))
  elif filename is None:
    common.fatal("read_yaml: have no idea what to do") # pragma: no cover -- sanity check
  elif "package:" in filename:
    package = '.'.join(__name__.split('.')[:-1])
    with resources.open_text(package,filename.replace("package:","")) as fid:
      return read_yaml(string=fid.read())
  else:
    if not os.path.isfile(filename):
      if common.LOGGING or common.VERBOSE:
        print("YAML file %s does not exist" % filename) # pragma: no cover -- too hard to test to bother
      return None
    try:
      yaml_data = Box().from_yaml(filename=filename,default_box=True,box_dots=True,default_box_none_transform=False)
    except:
      common.fatal("Cannot read YAML from %s: %s " % (filename,str(sys.exc_info()[1])))

  if common.LOGGING or common.VERBOSE:
    print("Read YAML data from %s" % (filename or "string"))

  data.unroll_dots(yaml_data)
  return yaml_data

def include_defaults(topo: Box, fname: str) -> None:
  defaults = read_yaml(fname)
  if defaults:
    topo.input.append(fname)
    topo.defaults = defaults + topo.defaults

def load(fname: str , local_defaults: str, sys_defaults: str) ->Box:
  topology = read_yaml(fname)
  if topology is None:
    common.fatal('Cannot read topology file: %s' % sys.exc_info()[0]) # pragma: no cover -- sanity check, getting here would be hard
  assert topology is not None
  topology.input = [ fname ]
  if not 'includes' in topology:
    topology.includes = [ 'defaults', 'global_defaults' ]
  if not isinstance(topology.includes,list):
    common.error( \
      "Topology 'includes' element (if present) should be a list", \
      category=common.IncorrectValue,module="topology")
    topology.includes = []

  if 'defaults' in topology.includes:
    if local_defaults:
      include_defaults(topology,local_defaults)
    else:
      local_defaults = os.path.dirname(os.path.abspath(fname))+"/topology-defaults.yml"
      user_defaults  = os.path.expanduser('~/topology-defaults.yml')
      if os.path.isfile(local_defaults):
        include_defaults(topology,local_defaults)
      elif os.path.isfile(user_defaults):
        include_defaults(topology,user_defaults)

  if sys_defaults and 'global_defaults' in topology.includes:
    include_defaults(topology,sys_defaults)

  return topology

def add_cli_args(topo: Box, args: argparse.Namespace) -> None:
  if args.device:
    topo.defaults.device = args.device

  if args.provider:
    topo.provider = args.provider

  if args.settings:
    for s in args.settings:
      if not "=" in s:
        common.error("Invalid CLI setting %s, should be in format key=value" % s)
      (k,v) = s.split("=")
      if '.' in k:
        try:
          data.set_dots(topo,k.split('.'),v)
        except TypeError as ex:
          if 'nodes.' in k:
            common.error(
              f'Cannot set {k}:\n... nodes element must be a dictionary if you want to set values via CLI arguments',
              common.IncorrectValue,
              'cli')
          elif 'links.' in k:
            common.error(
              f'Cannot set link value {k} through CLI arguments',
              common.IncorrectValue,
              'cli')
          else:
            common.fatal(f"Cannot set topology value {k}\n... {ex}")
        except Exception as ex:
          common.fatal(f"Cannot set topology value {k}\n... {ex}")
      else:
        topo[k] = v




"""
Routing Protocol utility functions:

* network_type: set IGP network type on a link. Used by OSPF and IS-IS
* routing_af: set routing protocol address families for a node
* external: return True is an interface is an external interface, and removes IGP-related parameters from the interface
* passive: set IGP 'passive' flag on an interface
"""

from box import Box
import typing
import netaddr

from .. import common
from .. import addressing
from ..data import get_from_box

# Build routing protocol address families
#
# * If the address families are not set, calculate them based on interface address families
# * Otherwise parse and validate the AF attribute
#
def routing_af(node: Box, proto: str) -> None:
  if 'af' in node[proto]:               # Is the AF attribute set for the routing protocol?
    if isinstance(node[proto].af,list): # Turn a list of address families into a dictionary
      node[proto].af = { af: True for af in node[proto].af }

    if node[proto].af is None:
      node[proto].pop('af',None)
    else:
      if not isinstance(node[proto].af,dict):
        common.error(
          f'af attribute for {proto} on node {node.name} has to be a list or a dictionary',
          common.IncorrectValue,
          proto)
        return

      for proto_af in node[proto].af.keys():
        if not proto_af in ('ipv4','ipv6'):
          common.error(
            f'Routing protocol address family has to be ipv4 and/or ipv6: {proto} on {node.name}',
            common.IncorrectValue,
            proto)
  
  if not 'af' in node[proto]:           # No configured AF attribute, calculate it
    for af in ['ipv4','ipv6']:
      if af in node.loopback:           # Address family enabled on loopback?
        node[proto].af[af] = True       # ... we need it in the routing protocol
        continue

      for l in node.get('interfaces',[]):              # Scan all interfaces
        if af in l and proto in l and not 'vrf' in l:  # Do we have AF enabled on any global interface?
          node[proto].af[af] = True                    # Found it - we need it the module
          continue

  for af in ['ipv4','ipv6']:              # Remove unused address families
    if not node[proto].af.get(af,False):
      node[proto].af.pop(af,False)

# Set network type for an interface:
#
# * If the network type is specified, validate it against a list of allowed network types
# * Otherwise, set network type to P2P if the interface has two nodes attached to it
#
def network_type(
      intf: Box,
      proto: str,
      allowed: typing.List[str] = ['point-to-point'],
      p2p: str = 'point-to-point') -> typing.Optional[str]:
  if 'network_type' in intf[proto]:                 # Did the user specify network type? 
    if not intf[proto].network_type:                # ... she did and she wants it gone
      intf.proto.pop('network_type')
    else:
      if intf[proto].network_type not in allowed:   # ... did she specify a valid value?
        return(f"Invalid {proto} network type {intf[proto].network_type}")
  elif len(intf.get('neighbors',[])) == 1:
    intf[proto].network_type = p2p                  # Network type not specified, set it for P2P links

  return None

# Remove routing protocol data from an interface with "external" role
#
def external(intf: Box, proto: str) -> bool:
  if intf.get('role','') == "external":
    intf.pop(proto,None)
    return True

  if proto in intf and isinstance(intf[proto],bool) and not intf[proto]:        # Disable IGP on an interface by setting it to False
    intf.pop(proto,None)
    return True

  return False

# IGP passive interfaces: stub link type or stub/passive role
#
def passive(intf: Box, proto: str) -> bool:
  if not 'passive' in intf[proto]:
    intf[proto].passive = intf.type == "stub" or intf.get('role',"") in ["stub","passive"]
  else:
    intf[proto].passive = bool(intf[proto].passive)
  return intf[proto].passive

# Create router ID if needed
#
def router_id(node: Box, proto: str, pools: Box) -> None:
  if 'router_id' in node.get(proto,{}):       # User-configured per-protocol router ID, get out of here
    try:
      node[proto].router_id = str(netaddr.IPAddress(node[proto].router_id).ipv4())
    except Exception as ex:
      common.error(
        f'{proto} router_id "{node[proto].router_id}" specified on node {node.name} is not an IPv4 address',
        common.IncorrectValue,
        proto)
    return

  if 'router_id' in node:                     # Node as configured router ID, copy it and get out
    try:
      node.router_id = str(netaddr.IPAddress(node.router_id).ipv4())
      node[proto].router_id = node.router_id
    except Exception as ex:
      common.error(
        f'router_id "{node.router_id}" specified on node {node.name} is not an IPv4 address',
        common.IncorrectValue,
        proto)
    return

  if 'ipv4' in node.get('loopback',{}):       # Do we have IPv4 address on the loopback? If so, use it as router ID
    node[proto].router_id = str(netaddr.IPNetwork(node.loopback.ipv4).ip)
    return

  if not pools.router_id:
    common.error(
      f'Cannot create a router ID for protocol {proto} on node {node.name}: router_id addressing pool is not defined',
      common.MissingValue,
      proto)
    return

  pfx = addressing.get(pools,['router_id'],node.id)
  if not pfx:
    common.error(
      f'Cannot create a router ID prefix from router_id pool for protocol {proto} on node {node.name}',
      common.IncorrectValue,
      proto)
    return

  if not pfx.get('ipv4',None):
    common.error(
      f'router_id pool did not return a usable IPv4 address to use as router ID for protocol {proto} on node {node.name}',
      common.IncorrectValue,
      proto)
    return

  node.router_id = str(netaddr.IPNetwork(pfx['ipv4']).ip)
  node[proto].router_id = node.router_id

#
# remove_vrf_interfaces -- remove interfaces in a VRF from a routing process that is not VRF-aware
#
def remove_vrf_interfaces(node: Box, proto: str) -> None:
  for l in node.interfaces:
    if proto in l and 'vrf' in l:
      l.pop(proto,None)

#
# build_vrf_interface_list -- copy VRF interfaces into VRF definition
#
def build_vrf_interface_list(node: Box, proto: str, topology: Box) -> None:
  for l in node.interfaces:
    if proto in l and 'vrf' in l:
      if not 'interfaces' in node.vrfs[l.vrf][proto]:                       # Start with an empty interface list
        node.vrfs[l.vrf][proto].interfaces = []
      if not 'active' in node.vrfs[l.vrf][proto]:                           # Assume there are no IGP neighbors in this VRF
        node.vrfs[l.vrf][proto].active = False
      node.vrfs[l.vrf][proto] = node[proto] + node.vrfs[l.vrf][proto]       # Add node IGP parameters to VRF IGP parameters
      node.vrfs[l.vrf][proto].interfaces.append(Box(l))                     # Append a copy of the interface data
      l.pop(proto,None)                                                     # ... and remove global IGP parameters from interface
                                                                            # Next we need to find if the VRF instance of IGP matters
      for neighbor in l.neighbors:                                          # ... iterate over the list of neighbors
        n_data = topology.nodes[neighbor.node]
        if proto in n_data.get('module',[]):                                # ... and check if at least one of them uses the IGP
          node.vrfs[l.vrf][proto].active = True
                                                                            # Cleanup IGP data
  for vdata in node.get('vrfs',{}).values():                                # ... iterate over the list of VRFs
    if not get_from_box(vdata,f'{proto}.active'):                           # ... and if there's no record of active IGP neighbors
      vdata.pop(proto,None)                                                 # ... remove the VRF IGP instance

#
# remove_unused_igp -- remove IGP module if it's not configured on any interface
#
def remove_unused_igp(node: Box, proto: str) -> None:
  if not any(proto in ifdata for ifdata in node.interfaces):                # Is protocol configured on any non-loopback interface?
    node.pop(proto,None)                                                    # ... no, remove protocol data from node

  if proto in node and 'af' in node[proto] and node[proto].af:              # Is at least one global AF active for the protocol?
    return                                                                  # ... OK, we're good

  for vdata in node.get('vrfs',{}).values():                                # Is protocol active in at least one VRF?
    if proto in vdata:
      return                                                                # ... OK, we're good

  node.module = [ m for m in node.module if m != proto ]                    # Makes no sense to keep it, remove the config module



#
# netlab config command
#
# Deploy custom configuration template to network devices
#
import typing
import argparse
import os
import glob

from . import common_parse_args
from . import ansible
from .. import common

#
# CLI parser for 'netlab config' command
#
def custom_config_parse(args: typing.List[str]) -> typing.Tuple[argparse.Namespace, typing.List[str]]:
  parser = argparse.ArgumentParser(
    prog='netlab config',
    description='Deploy custom configuration template',
    epilog='All other arguments are passed directly to ansible-playbook')
  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='store_true',
    help='Verbose logging')
  parser.add_argument(
    dest='template', action='store',
    help='Configuration template (or a family of templates)')

  return parser.parse_known_args(args)

def run(cli_args: typing.List[str]) -> None:
  (args,rest) = custom_config_parse(cli_args)

  if args.verbose:
    rest = ['-v'] + rest

  if args.template != '-':
    if os.path.exists(args.template) or \
       os.path.exists(args.template+'.j2') or \
       glob.glob(args.template+'.*.j2'):
      rest = ['-e','config='+args.template] + rest
    else:
      common.fatal('Cannot find specified Jinja2 template: %s' % args.template,'config')

  if args.verbose:
    print('Ansible playbook args: %s' % rest)
  ansible.playbook('config.ansible',rest)



"""setup.py file."""
from setuptools import setup, find_packages
import netsim
import sys

with open("requirements.txt", "r") as fs:
  reqs = [r for r in fs.read().splitlines() if (len(r) > 0 and not r.startswith("#"))]

if sys.version_info < (3, 7):
  raise RuntimeError("This package requires Python 3.7+")

setup(
  name="netsim-tools",
  version=netsim.__version__,
  packages=find_packages(),
  author="Ivan Pepelnjak",
  author_email="ip@ipspace.net",
  description="CLI-based Virtual Networking Lab Abstraction Layer",
  classifiers=[
    "Topic :: Utilities",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.9",
    "Operating System :: POSIX :: Linux",
    "Operating System :: MacOS",
  ],
  url="https://github.com/ipspace/netsim-tools",
  include_package_data=True,
  setup_requires=["wheel"],
  python_requires='>=3.7',  # Due to e.g. 'capture_output' in subprocess.run
  install_requires=reqs,
  entry_points={
    "console_scripts": ["netlab=netsim.cli:lab_commands"]
  },
  package_data = {
    "netsim": ["templates/*", "ansible/*", "extra/*"]
  }
)



#
# Create Ansible inventory
#
import typing

import yaml
import os
from box import Box

from .. import common
from . import _TopologyOutput

forwarded_port_name = { 'ssh': 'ansible_port', }

def provider_inventory_settings(node: Box, defaults: Box) -> None:
  p_data = defaults.providers[defaults.provider]
  if not p_data:
    return        # pragma: no cover -- won't create an extra test case just to cover the "do nothing" scenario

  if 'inventory' in p_data:
    for k,v in p_data['inventory'].items():
      node[k] = v

  if 'inventory_port_map' in p_data and 'forwarded' in p_data:
    for k,v in p_data['inventory_port_map'].items():
      if k in p_data['forwarded']:
        node[v] = p_data['forwarded'][k] + node['id']

topo_to_host = { 'mgmt.ipv4': 'ansible_host', 'hostname': 'ansible_host', 'id': 'id' }
topo_to_host_skip = [ 'name','device' ]

def ansible_inventory_host(node: Box, defaults: Box) -> Box:
  host = Box({})
  for (node_key,inv_key) in topo_to_host.items():
    if "." in node_key:
      value = node[node_key]
    else:
      value = node.get(node_key,None)
    if value:
      host[inv_key] = value

  for (k,v) in node.items():
    if not k in topo_to_host_skip:
      host[k] = v

  provider_inventory_settings(host,defaults)
  return host

def create(nodes: Box, groups: Box, defaults: Box, addressing: typing.Optional[Box] = None) -> Box:
  inventory = Box({},default_box=True,box_dots=True)

  inventory.all.vars.netlab_provider = defaults.provider
  inventory.modules.hosts = {}
  inventory.custom_configs.hosts = {}

  if 'module' in defaults:
    inventory.modules.vars.netlab_module = defaults.module

  if addressing:
    inventory.all.vars.pools = addressing
    for name,pool in inventory.all.vars.pools.items():
      for k in list(pool.keys()):
        if ('_pfx' in k) or ('_eui' in k):
          del pool[k]

  for name,node in nodes.items():
    group = node.get('device','all')
    inventory[group].hosts[name] = ansible_inventory_host(node,defaults)

    if 'module' in node:
      inventory.modules.hosts[name] = {}

    if 'config' in node:
      inventory.custom_configs.hosts[name] = {}

  if 'devices' in defaults:
    for group in inventory.keys():
      if group in defaults.devices:
        devdata = defaults.devices[group]
        group_vars = devdata.group_vars + devdata[defaults.provider].group_vars
        if group_vars:
          inventory[group]['vars'] = group_vars

  for gname,gdata in groups.items():
    if not gname in inventory:
      inventory[gname] = { 'hosts': {} }

    if 'vars' in gdata:
      inventory[gname].vars = inventory[gname].get('vars',{}) + gdata.vars

    if 'members' in gdata:
      for m in gdata.members:
        if m in nodes:
          if not m in inventory[gname].hosts:
            inventory[gname].hosts[m] = {}
        elif m in groups:
          inventory[gname].children[m] = {}

  return inventory

def dump(data: Box) -> None:
  print("Ansible inventory data")
  print("===============================")
  inventory = create(data.nodes,data.get('groups',{}),data.defaults)
  print(inventory.to_yaml())

def write_yaml(data: Box, fname: str, header: str) -> None:
  dirname = os.path.dirname(fname)
  if dirname and not os.path.exists(dirname):
    os.makedirs(dirname)

  with open(fname,"w") as output:
    output.write(header)
    if callable(getattr(data,"to_yaml",None)):
      output.write(data.to_yaml())
    else:                            # pragma: no cover -- this should never happen as we're using Box, but just in case...
      output.write(yaml.dump(data))
    output.close()

min_inventory_data = [ 'id','ansible_host','ansible_port' ]

def ansible_inventory(data: Box, fname: typing.Optional[str] = 'hosts.yml', hostvars: typing.Optional[str] = 'dirs') -> None:
  inventory = create(data['nodes'],data.get('groups',{}),data.get('defaults',{}),data.get('addressing',{}))

#  import ipdb; ipdb.set_trace()
  header = "# Ansible inventory created from %s\n#\n---\n" % data.get('input','<unknown>')

  if not fname:
    fname = 'hosts.yml'
  if not hostvars:
    hostvars = "dirs"

  if hostvars == "min":
    write_yaml(inventory,fname,header)
    print("Created single-file Ansible inventory %s" % fname)
  else:
    for g in inventory.keys():
      gvars = inventory[g].pop('vars',None)
      if gvars:
        write_yaml(gvars,'group_vars/%s/topology.yml' % g,header)
        if not common.QUIET:
          print("Created group_vars for %s" % g)

      if 'hosts' in inventory[g]:
        hosts = inventory[g]['hosts']
        for h in hosts.keys():
          if not hosts[h]:
            continue
          min_host = Box({})
          vars_host = Box({})
          for item in hosts[h].keys():
            if item in min_inventory_data:
              min_host[item] = hosts[h][item]
            else:
              vars_host[item] = hosts[h][item]

          write_yaml(vars_host,'host_vars/%s/topology.yml' % h,header)
          if not common.QUIET:
            print("Created host_vars for %s" % h)
          hosts[h] = min_host

    write_yaml(inventory,fname,header)
    print("Created minimized Ansible inventory %s" % fname)

def ansible_config(config_file: typing.Union[str,None] = 'ansible.cfg', inventory_file: typing.Union[str,None] = 'hosts.yml') -> None:
  if not config_file:
    config_file = 'ansible.cfg'
  if not inventory_file:
    inventory_file = 'hosts.yml'

  with open(config_file,"w") as output:
    output.write(common.template('ansible.cfg.j2',{ 'inventory': inventory_file or 'hosts.yml' },'templates'))
    output.close()
    if not common.QUIET:
      print("Created Ansible configuration file: %s" % config_file)

class AnsibleInventory(_TopologyOutput):

  def write(self, topology: Box) -> None:
    hostfile = self.settings.hostfile or 'hosts.yml'
    configfile = self.settings.configfile or 'ansible.cfg'
    output_format = None

    if hasattr(self,'filenames'):
      hostfile = self.filenames[0]
      if len(self.filenames) > 1:
        configfile = self.filenames[1]
      if len(self.filenames) > 2:
        common.error('Extra output filename(s) ignored: %s' % str(self.filenames[2:]),common.IncorrectValue,'ansible')

    if self.format:
      output_format = self.format[0]

    ansible_inventory(topology,hostfile,output_format)
    ansible_config(configfile,hostfile)



#
# netlab create command
#
# Creates virtualization provider configuration and automation inventory from
# the specified topology
#
import argparse
import typing
import textwrap
from box import Box

from . import common_parse_args, topology_parse_args, load_topology
from .. import read_topology,augment,common
from ..outputs import _TopologyOutput

#
# CLI parser for create-topology script
#
def create_topology_parse(
      args: typing.List[str],
      cmd: str,
      description: str,
      extra_args: typing.Optional[argparse.ArgumentParser]) -> argparse.Namespace:
  if cmd != 'create':
    epilog = ""
  else:
    epilog = textwrap.dedent('''
      output files created when no output is specified:

        * Transformed topology snapshot in netlab.snapshot.yml
        * Virtualization provider file with provider-specific filename
          (Vagrantfile or clab.yml)
        * Ansible inventory file (hosts.yml) and configuration (ansible.cfg)

      For a complete list of output formats please consult the documentation
    ''')
  parents = [ common_parse_args(True), topology_parse_args() ]
  if extra_args:
    parents.append(extra_args)
  parser = argparse.ArgumentParser(
    parents=parents,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    prog=f"netlab {cmd}",
    description=description,
    epilog=epilog)

  parser.add_argument(
    dest='topology', action='store', nargs='?',
    type=argparse.FileType('r'),
    default='topology.yml',
    help='Topology file (default: topology.yml)')

  if cmd == 'create':
    parser.add_argument('-o','--output',dest='output', action='append',help='Output format(s): format:option=filename')
    parser.add_argument('--devices',dest='devices', action='store_true',help='Create provider configuration file and netsim-devices.yml')

  return parser.parse_args(args)

def run(cli_args: typing.List[str],
        cli_command: str = 'create',
        cli_describe: str = 'Create provider- and automation configuration files',
        cli_extra_args: typing.Optional[argparse.ArgumentParser] = None ) -> Box:
  args = create_topology_parse(cli_args, cli_command, cli_describe, cli_extra_args)
  if not 'output' in args:
    args.output = None
  if not 'devices' in args:
    args.devices = None

  if not args.output:
    args.output = ['provider','yaml=netlab.snapshot.yml']
    args.output.append('devices' if args.devices else 'ansible:dirs')
  elif args.devices:
    common.error('--output and --devices flags are mutually exclusive',common.IncorrectValue,'create')

  topology = load_topology(args)
  augment.main.transform(topology)
  common.exit_on_error()

  for output_format in args.output:
    output_module = _TopologyOutput.load(output_format,topology.defaults.outputs[output_format])
    if output_module:
      output_module.write(topology)
    else:
      common.error('Unknown output format %s' % output_format,common.IncorrectValue,'create')

  return topology



'''
Add group support. This module handles group-related data structures and
data transformation. The end result is a dictionary of groups in 'groups'
top-level element. That dictionary is merged from global- and node-level parameters
'''

import typing

from box import Box

from .. import common
from .. import data
from . import nodes

group_attr = [ 'members','vars','config','node_data','device','module' ]

'''
Return members of the specified group. Recurse through child groups if needed
'''
def group_members(topology: Box, group: str, count: int = 0) -> list:
  members: typing.List[str] = []
  if not group in topology.groups:  # pragma: no cover (just-in case catch, impossible to get here)
    common.error(
      f'Internal error: unknown group {group}',
      common.IncorrectValue,
      'groups')                    
    return []                      

  if count > 99:                    # pragma: no cover (impossible to get here, recursive groups are checked elsewhere)
    common.fatal(
      'Recursive group definition, aborting',
      'groups')                    

  for m in topology.groups[group].members:
    if m in topology.nodes:
      members = members + [ m ]
    if m in topology.groups:
      members = members + group_members(topology,m,count + 1)

  return members

'''
Check validity of 'groups' data structure
'''
def check_group_data_structure(topology: Box) -> None:
  if not 'groups' in topology:
    topology.groups = Box({},default_box=True,box_dots=True)

  if not isinstance(topology.groups,dict):
    common.error('Groups topology-level element should be a dictionary of groups')
    return

  '''
  Transform group-as-list into group-as-dictionary
  '''
  for grp in topology.groups.keys():
    if isinstance(topology.groups[grp],list):
      topology.groups[grp] = { 'members': topology.groups[grp] }
    if grp in topology.nodes:
      common.error(
        f"group {grp} is also a node name. I can't deal with that level of confusion",
        common.IncorrectValue,
        'groups')

  '''
  Sanity checks on global group data
  '''
  for grp,gdata in topology.groups.items():
    if not isinstance(gdata,Box):
      common.error('Group definition should be a dictionary: %s' % grp)
      continue

    if not 'members' in gdata:
      gdata.members = []

    if grp == 'all' and gdata.members:
      common.error('Group "all" should not have explicit members')

    if 'vars' in gdata:
      if not isinstance(gdata.vars,dict):
        common.error('Group variables must be a dictionary: %s' % grp)

    for attr in ('config','module'):
      if attr in gdata:
        data.must_be_list(gdata,attr,f'groups.{grp}')

    for attr in ('device'):
      data.must_be_string(gdata,attr,f'groups.{grp}')

    for k in gdata.keys():
      if not k in group_attr:
        common.error('Unknown attribute "%s" in group %s' % (k,grp))

    if not isinstance(gdata.members,list):
      common.error('Group members must be a list of nodes or groups: %s' % grp)
      continue

    for n in gdata.members:
      if not n in topology.nodes and not n in topology.groups:
        common.error('Member %s of group %s is not a valid node or group name' % (n,grp))


'''
Add node-level group settings to global groups
'''
def add_node_level_groups(topology: Box) -> None:
  for name,n in topology.nodes.items():
    if not 'group' in n:
      continue

    if isinstance(n.group,str):                           # Node group should be a list
      n.group = [ n.group ]                               # Convert a string value into a single-element list

    for grpname in n.group:
      if not grpname in topology.groups:
        topology.groups[grpname] = { 'members': [] }      # Create an empty new group if needed

      if not name in topology.groups[grpname].members:  # Node not yet in the target group
        topology.groups[grpname].members.append(name)   # Add node to the end of the member list

'''
Check recursive group definitions
'''

def check_recursive_chain(topology: Box, chain: list, group: str) -> typing.Optional[list]:
  if not group in topology.groups: # pragma: no cover (if we ever get here we're seriously messed up)
    common.fatal(
      'Internal error: unknown group in check_recursive_chain') 
    return None  

  chain = chain + [ group ]
  for m in topology.groups[group].members:
    if m in chain:
      chain = chain + [ m ]
      common.error(f'Recursive group definition chain {chain}', common.IncorrectValue, 'groups')
      return chain
    if m in topology.groups:
      if check_recursive_chain(topology,chain,m):
        return chain

  return None

def check_recursive_groups(topology : Box) -> None:
  for gname in topology.groups.keys():
    if check_recursive_chain(topology,[],gname):
      return

def reverse_topsort(topology: Box) -> list:
  group_copy = Box(topology.groups)            # Make a copy of the group dictionary
  sort_list: typing.List[str] = []
  while group_copy:                            # Keep iterating until we got all groups in order
    for g in sorted(group_copy.keys()):        # Iterate over remaining groups
      OK_to_add = True                         # ... sort them by name to have consistent results
      for m in group_copy[g].members:          # Now go check the group members
        if m in group_copy:                    # ... if a member of this group is still in groups we're not yet ready
          OK_to_add = False                    # ... to add it to sorted list
      if OK_to_add:                            # Have all child groups already been removed from the group dictionary?
        sort_list = sort_list + [ g ]          # ... if so, we're good to go, add current group to sorted list
        group_copy.pop(g)                      # ... and remove it from dictionary of remaining groups

  return sort_list

'''
Copy group-level module or device setting into node data
'''
def copy_group_device_module(topology: Box) -> None:
  for grp in reverse_topsort(topology):
    gdata = topology.groups[grp]
    if 'device' in gdata or 'module' in gdata:
      g_members = group_members(topology,grp)
      if g_members:
        for name,ndata in topology.nodes.items():
          if name in g_members:
            for attr in ('device','module'):
              if attr in gdata and not attr in ndata:
                ndata[attr] = gdata[attr]
      else:
        common.error(
          f'Cannot use "module" or "device" attribute on in group {grp} that has no direct or indirect members',
          common.IncorrectValue,
          'groups')

'''
Copy node data from group into group members
'''
def copy_group_node_data(topology: Box) -> None:
  for grp in reverse_topsort(topology):
    gdata = topology.groups[grp]
    if 'node_data' in gdata:
      node_data_keys  = set(gdata.node_data.keys())                   # You cannot set device type or node modules with
      node_data_wrong = False                                         # ... node data, those attributes have to be set
      for k in ('module','device'):                                   # ... at group level
        if k in node_data_keys:
          common.error(
            f'Cannot use attribute {k} in node_data in group {grp}, set it as a group attribute',
            common.IncorrectValue,
            'groups')
          node_data_wrong = True
      if node_data_wrong:
        continue

      g_members = group_members(topology,grp)
      for name,ndata in topology.nodes.items():
        if name in g_members:
          for k,v in gdata.node_data.items():   # Have to go one level deeper, changing ndata value wouldn't work
            if not k in ndata:
              ndata[k] = v
            if isinstance(ndata[k],dict):
              if isinstance(v,dict):
                ndata[k] = ndata[k] + v
              else:
                common.error(
                  f'Cannot merge non-dictionary node_data {k} from group {grp} into node {ndata.name}',
                  common.IncorrectValue,
                  'groups')

#
# init_groups:
#
# * Check and adjust group data structures
# * Check recursive groups
# * Add nodes to groups based on node 'group' attribute
#
# Please note that check_group_data_structure creates 'groups' element if needed
# and 'adjust_groups' deletes it if there are no groups in the topology.
#
def init_groups(topology: Box) -> None:
  check_group_data_structure(topology)
  add_node_level_groups(topology)
  common.exit_on_error()

  check_recursive_groups(topology)
  common.exit_on_error()

  copy_group_device_module(topology)

#
# adjust_groups:
#
# * Copy group data into nodes
#
# We have to split the group initialization code into two phases
# due to BGP auto groups.
#
def adjust_groups(topology: Box) -> None:
  copy_group_node_data(topology)
  common.exit_on_error()

  '''
  Finally, remove 'groups' topology element if it's not needed
  '''
  if not topology.groups:
    del topology['groups']

'''
Copy custom config templates from groups into nodes
'''
def node_config_templates(topology: Box) -> None:
  if not 'groups' in topology:
    return

  '''
  Phase 1 - merge group config templates into nodes

  Traverse groups from more-specific to less specific, pushing config templates in front of
  the accumulated config list.

  End result: config templates sorted from less specific through more specific groups, ending
  with node templates.
  '''

  for group_name in reverse_topsort(topology):
    if not 'config' in topology.groups[group_name]:
      continue

    data.must_be_list(topology.groups[group_name],'config',f'groups.{group_name}')
    g_members = group_members(topology,group_name)
    for name,ndata in topology.nodes.items():
      if name in g_members or group_name == 'all':
        if not data.must_be_list(ndata,'config',f'nodes.{name}') is None:
          ndata.config = topology.groups[group_name].config + ndata.config

  '''
  Phase 2 - cleanup

  * Remove 'config' attributes from groups (just in case, they are no longer needed anyway)
  * Process node 'config' attributes to honor 'removal of prior templates' requests
  '''

  for group_name in topology.groups:
    topology.groups[group_name].pop('config',None)

  for name,node in topology.nodes.items():
    if not 'config' in node:
      continue

    config_list: typing.List[str] = []
    for c in node.config:
      if c == '-':                 # Remove all prior templates
        config_list = []
      elif c[0] == '-':
        config_list = [ t for t in config_list if t != c [1:] ]
      else:
        config_list = config_list + [ c ]

    if config_list:
      node.config = config_list
    else:
      node.pop('config',None)



"""
Device utility functions
"""

import typing

from box import Box

from .. import common

"""
Get generic device attribute:

* Use node.device to find device used by the current node
* Use defaults.provider (future: node data) to find the provider
* Fetch required data using the following inheritance rules:

  * If the provider data is not a dictionary, return that (no merge)
  * If the provider data is a dictionary, but the device data is not, return provider data (override)
  * Return a merge of both dictionaries
"""
def get_device_attribute(node: Box, attr: str, defaults: Box) -> typing.Optional[typing.Any]:
  devtype  = node.device
  provider = defaults.provider

  if not devtype in defaults.devices:    # pragma: no cover
    common.fatal(f'Internal error: call to get_device_attribute with unknown device {devtype}')
    return None

  pvalue = None
  devdata = defaults.devices[devtype]

  if provider in devdata:                # Does this device have attributes for current provider?
  	if attr in devdata[provider]:        # Including the attribute we're looking for?
  	  pvalue = devdata[provider][attr]   # Get the provider value
  	  if not isinstance(pvalue,dict):    # Provider-specific value is not a dictionary
  	  	return pvalue                    # No chance of merging, return it

  value = devdata.get(attr,None)         # Non-specific device data
  if isinstance(value,Box) and isinstance(pvalue,Box):
    return value+pvalue                  # Return merged dictionaries

  if attr in devdata[provider]:          # Do we have a provider-specific value?
    return pvalue                        # Provider-specific dictionary overriding non-dictionary device value

  return value                           # Return whatever the device value is

"""
Get device feature flags -- uses get_device_attribute but returns a Box to keep mypy happy
"""
def get_device_features(node: Box, defaults: Box) -> Box:
  features = get_device_attribute(node,'features',defaults)
  if not features:
    return Box({},default_box=True,box_dots=True)

  if not isinstance(features,Box):
    common.fatal('Device features for device type {node.device} should be a dictionary')
    return Box({})

  return features

"""
Get all device data for current provider
"""
def get_provider_data(node: Box, defaults: Box) -> Box:
  devtype  = node.device
  provider = defaults.provider

  if not devtype in defaults.devices:
    common.fatal(f'Internal error: call to get_provider_data with unknown device {devtype}')

  return defaults.devices[devtype].get(provider,{})

"""
Get consolidated device data
"""
def get_consolidated_device_data(node: Box, defaults: Box) -> Box:
  devtype  = node.device
  provider = defaults.provider

  if not devtype in defaults.devices:
    common.fatal(f'Internal error: call to get_provider_data with unknown device {devtype}')

  data = defaults.devices[devtype] + defaults.devices[devtype].get(provider,{})
  for p in defaults.providers.keys():
    data.pop(p,None)

  return data

"""
Get node provider -- currently returns the default provider, but we'll do fun stuff pretty soon ;)
"""
def get_provider(node: Box, defaults: Box) -> str:
  return defaults.provider

def augment_device_settings(topology: Box) -> None:
  pass



#!/usr/bin/env python3
#
# Main CLI entry point
#

import sys
import importlib
import argparse
import os
import shutil
import typing

from box import Box

from . import usage
from .. import augment, common, read_topology

def common_parse_args(debugging: bool = False) -> argparse.ArgumentParser:
  parser = argparse.ArgumentParser(description='Common argument parsing',add_help=False)
  parser.add_argument('--log', dest='logging', action='store_true',
                  help='Enable basic logging')
  parser.add_argument('-q','--quiet', dest='quiet', action='store_true',
                  help='Report only major errors')
  parser.add_argument('-v','--verbose', dest='verbose', action='count', default = 0,
                  help='Verbose logging (add multiple flags for increased verbosity)')
  parser.add_argument('--warning', dest='warning', action='store_true',help=argparse.SUPPRESS)
  parser.add_argument('--raise_on_error', dest='raise_on_error', action='store_true',help=argparse.SUPPRESS)
  if debugging:
    parser.add_argument('--debug', dest='debug', action='store_true',
                    help='Debugging (might not execute external commands)')

  return parser

def topology_parse_args() -> argparse.ArgumentParser:
  parser = argparse.ArgumentParser(description='Common topology arguments',add_help=False)
  parser.add_argument('--defaults', dest='defaults', action='store',
                  help='Local topology defaults file')
  parser.add_argument('-d','--device', dest='device', action='store', help='Default device type')
  parser.add_argument('-p','--provider', dest='provider', action='store',help='Override virtualization provider')
  parser.add_argument('-s','--set',dest='settings', action='append',help='Additional parameters added to topology file')
  return parser

#
# Common file/directory cleanup routine, used by collect/clab/down
#

def fs_cleanup(filelist: typing.List[str], verbose: bool = False) -> None:
  for fname in filelist:
    if os.path.isdir(fname):
      if verbose:
        print(f"... removing directory tree {fname}")
      try:
        shutil.rmtree(fname)
      except Exception as ex:
        common.fatal(f"Cannot clean up directory {fname}: {ex}")
    elif os.path.exists(fname):
      if verbose:
        print(f"... removing {fname}")
      try:
        os.remove(fname)
      except Exception as ex:
        common.fatal(f"Cannot remove {fname}: {ex}")

# Common topology loader (used by create and down)

def load_topology(args: argparse.Namespace) -> Box:
  common.set_logging_flags(args)
  topology = read_topology.load(args.topology.name,args.defaults,"package:topology-defaults.yml")

  if args.settings or args.device or args.provider:
    topology.nodes = augment.nodes.create_node_dict(topology.nodes)
    read_topology.add_cli_args(topology,args)

  common.exit_on_error()
  return topology

# Snapshot-or-topology loader (used by down)

def load_snapshot_or_topology(args: argparse.Namespace) -> typing.Optional[Box]:
  common.set_logging_flags(args)
  if args.device or args.provider or args.settings:     # If we have -d, -p or -s flag
    if not args.topology:                               # ... then the user wants to use the topology file
      args.topology = 'topology.yml'                    # ... so let's set the default value if needed

  if args.topology:
    topology = load_topology(args)
    augment.main.transform(topology)
    common.exit_on_error()
    return topology
  else:
    args.snapshot = args.snapshot or 'netlab.snapshot.yml'
    return read_topology.read_yaml(filename=args.snapshot)

#
# Main command dispatcher
#
# The command to execute is the first CLI argument. We try to load a module with the same name
# and if the module load fails for whatever reason we generate an error and exit.
#
# However, the module load could fail due to coding errors within the module (not because the module
# does not exist), in which case the 'debug' command becomes useful - it loads the required module
# without the exception handling, resulting in a nice error printout.
#
# After we have the module that we hope will execute the desired command, we check whether it has
# a 'run' function, and if it does, we call it with the remaining arguments.
#

quick_commands = {
  'alias': lambda x: usage.print_usage('alias.txt')
}

def lab_commands() -> None:
  if len(sys.argv) < 2:
    usage.run([])
    sys.exit()

  arg_start = 2
  mod = None
  cmd = sys.argv[1]

  if cmd == 'debug':
    arg_start = 3
    mod = importlib.import_module("."+sys.argv[2],__name__)
  elif quick_commands.get(cmd,None):
    quick_commands[cmd](sys.argv[arg_start:])
    return
  else:
    try:
      mod = importlib.import_module("."+cmd,__name__)
    except Exception as ex:
      print( f"Error importing .{cmd},{__name__}: {ex}" )
      pass

  if mod and hasattr(mod,'run'):
    mod.run(sys.argv[arg_start:])   # type: ignore
    return

  print("Invalid CLI command: %s\n\nUse 'netlab usage' to get the list of valid commands" % cmd)
  sys.exit(1)



#
# netlab collect command
#
# Collect device configurations
#
import typing
import os
import argparse
import shutil
import subprocess

from . import common_parse_args,fs_cleanup
from . import ansible
from . import external_commands
from .. import common

#
# CLI parser for 'netlab collect' command
#
def initial_config_parse(args: typing.List[str]) -> typing.Tuple[argparse.Namespace, typing.List[str]]:
  parser = argparse.ArgumentParser(
    prog="netlab collect",
    description='Collect device configurations',
    epilog='All other arguments are passed directly to ansible-playbook')

  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='store_true',
    help='Verbose logging')
  parser.add_argument(
    '-q','--quiet',
    dest='quiet',
    action='store_true',
    help='Run Ansible playbook and tar with minimum output')
  parser.add_argument(
    '-o','--output',
    dest='output',
    action='store',
    nargs='?',
    default='config',
    help='Output directory (default: config)')
  parser.add_argument(
    '--tar',
    dest='tar',
    action='store',
    help='Create configuration tarball')
  parser.add_argument(
    '--cleanup',
    dest='cleanup',
    action='store_true',
    help='Clean up config directory and modified configuration file after creating tarball')
  return parser.parse_known_args(args)

def get_tarball_file(tarball: str) -> str:
  if not '.tar' in tarball:
    tarball = tarball + '.tar.gz'
  return tarball

def run(cli_args: typing.List[str]) -> None:
  (args,rest) = initial_config_parse(cli_args)
  common.set_logging_flags(args)

  fs_cleanup([ args.output ])
  try:
    os.mkdir(args.output)
  except Exception as ex:
    common.fatal(f"Cannot create output directory {{args.output}}: {{ex}}")

  if args.verbose:
    rest = ['-v'] + rest

  rest = ['-e','target='+args.output ] + rest

  if args.tar and not args.quiet:
    external_commands.print_step(1,"Collecting device configurations")

  if args.verbose:
    print("Ansible playbook args: %s" % rest)

  if args.quiet:
    os.environ["ANSIBLE_STDOUT_CALLBACK"] = "dense"

  ansible.playbook('collect-configs.ansible',rest)

  if args.tar:
    tarball = get_tarball_file(args.tar)
    if not args.quiet:
      external_commands.print_step(2,"Creating tarball",spacing = True)
    try:
      subprocess.check_call(['tar','cfz' if args.quiet else 'cvfz',tarball,args.output])
    except Exception as ex:
      common.fatal(f"Cannot start tar: {ex}")

    if args.cleanup:
      if not args.quiet:
        external_commands.print_step(3,"Cleanup config directory",spacing = True)
      fs_cleanup([ args.output ],args.verbose)
      print("... done")



#
# BGP neighbors transformation module to support a different underlay AS per node
#
# The core bgp module only supports a single AS per node, as that is the more
# common and simple way to operate BGP. Most vendors/devices can support
# multiple AS, for example in a topology that uses iBGP for an EVPN overlay and
# eBGP for the underlay.
#
# It is assumed the core bgp module provides the iBGP overlay peerings. This
# module adds an 'underlay_as' per node, and builds additional eBGP peerings
# based on that attribute.
#
# Changes:
# * Creates additional bgp.neighbors of type 'ebgp' with a 'local_as' attribute
#
import typing, netaddr
from box import Box

"""
Adds a custom bgp.underlay_as node attribute
"""
def init(topology: Box) -> None:
    # Allow users to specify a different AS to use for underlay peering (eBGP)
    topology.defaults.bgp.attributes.node.append('underlay_as')

def ebgp_neighbor(n: Box, asn: int, intf: Box, extra_data: dict) -> Box:
  ngb = Box(extra_data,default_box=True,box_dots=True)
  ngb.name = n.name
  ngb.type = 'ebgp'
  ngb["as"] = asn  # 'as' for backwards compatibility, would prefer 'peer_as'
  for af in ["ipv4","ipv6"]:
    if af in intf:
      if "unnumbered" in ngb and ngb.unnumbered == True:
        ngb[af] = True
      else:
        ngb[af] = str(netaddr.IPNetwork(intf[af]).ip)
  return ngb


"""
build_ebgp_sessions: augment BGP neighbors with ebgp peers
* EBGP sessions are established whenever two nodes on the same link have
  different underlay AS
"""
def build_ebgp_sessions(node: Box, topology: Box) -> None:

    #
    # eBGP sessions - iterate over all links, find adjacent nodes
    # in different AS numbers, and create eBGP neighbors; set 'local_as'
    ibgp_as = topology.bgp['as']
    node_as = node.bgp.underlay_as
    for l in node.get("interfaces",[]):
      for ngb_ifdata in l.get("neighbors",[]):
        ngb_name = ngb_ifdata.node
        neighbor = topology.nodes[ngb_name]
        if not "bgp" in neighbor or not "underlay_as" in neighbor.bgp:
          continue

        peer_as = neighbor.bgp.underlay_as

        if node_as!=peer_as:
          extra_data = Box({})
          extra_data.ifindex = l.ifindex
          if node_as != ibgp_as:
            extra_data.local_as = node_as
          if "unnumbered" in l:
            extra_data.unnumbered = True
            extra_data.local_if = l.ifname
          node.bgp.neighbors.append( ebgp_neighbor(neighbor,peer_as,ngb_ifdata,extra_data) )

def post_transform(topology: Box) -> None:
  for node in topology.nodes.values():
    if 'bgp' in node and 'underlay_as' in node.bgp:
        build_ebgp_sessions(node,topology)



#
# Dynamic virtualization provider framework
#
# Individual virtualization providers are defined in modules within this directory inheriting
# Provider class and replacing or augmenting its methods (most commonly, transform)
#

import platform
import subprocess
import os
import typing

# Related modules
from box import Box

from .. import common
from ..callback import Callback

class _Provider(Callback):
  def __init__(self, provider: str, data: Box) -> None:
    self.provider = provider
    if 'template' in data:
      self._default_template_name = data.template

  @classmethod
  def load(self, provider: str, data: Box) -> '_Provider':
    module_name = __name__+"."+provider
    obj = self.find_class(module_name)
    if obj:
      return obj(provider,data)
    else:
      return _Provider(provider,data)

  def get_template_path(self) -> str:
    return 'templates/provider/' + self.provider

  def get_output_name(self, fname: typing.Optional[str], topology: Box) -> str:
    if fname:
      return fname

    fname = topology.defaults.providers[self.provider].config
    if fname:
      return fname

    return "Vagrantfile"

  _default_template_name = "Vagrantfile.j2"

  def get_root_template(self) -> str:
    return self._default_template_name

  def node_image_version(self, topology: Box) -> None:
    for name,n in topology.nodes.items():
      if '.' in n.box:
        image_spec = n.box.split(':')
        n.box = image_spec[0]
        if len(image_spec) > 1:
          n.box_version = image_spec[1]

  def transform_node_images(self, topology: Box) -> None:
    pass

  def transform(self, topology: Box) -> None:
    self.transform_node_images(topology)
    if "processor" in topology.defaults:
      return
    else:
      processor_name = ""
      if platform.system() == "Windows":
        processor_name = platform.processor()
      elif platform.system() == "Darwin":
        processor_name = "intel"  # Assume Intel for MacOS
      elif platform.system() == "Linux":
        processor_name = str(subprocess.check_output("cat /proc/cpuinfo", shell=True).splitlines()[1].split()[2])
      topology.defaults.processor = processor_name

  def create(self, topology: Box, fname: typing.Optional[str]) -> None:
    self.transform(topology)
    fname = self.get_output_name(fname,topology)
    output = common.open_output_file(fname)
    output.write(common.template(self.get_root_template(),topology.to_dict(),self.get_template_path()))
    if fname != '-':
      common.close_output_file(output)
      print("Created provider configuration file: %s" % fname)
    else:
      output.write("\n")

  def post_start_lab(self, topology: Box) -> None:
    pass

  def pre_start_lab(self, topology: Box) -> None:
    pass

  def pre_stop_lab(self, topology: Box) -> None:
    pass

  def post_stop_lab(self, topology: Box) -> None:
    pass
